/*********************************************************************************
 *  SuperWaba Virtual Machine, version 5                                         *
 *  Copyright (C) 2000-2005 SuperWaba Ltda.                                      *
 *  All Rights Reserved                                                          *
 *                                                                               *
 *  This library and virtual machine is distributed in the hope that it will     *
 *  be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                         *
 *                                                                               *
 *  Please see the software license located at /license.txt for more details.    *
 *                                                                               *
 *  You should have received a copy of the License along with this software;     *
 *  if not, write to                                                             *
 *                                                                               *
 *     SuperWaba Ltda.                                                           *
 *     Rua Padre Leonel Franca, 480, sala 27A                                    *
 *     Instituto Genesis - PUC-RIO                                               *
 *     Rio de Janeiro / RJ - Brazil                                              *
 *     Cep: 22451-000                                                            *
 *     E-mail: licensing@superwaba.com.br                                        *
 *********************************************************************************/
/* $Id: SW_messages.c,v 1.15 2005/09/25 12:56:34 guich Exp $ */

#if HAVE_CONFIG_H
#include "config.h"
#endif

#if HAVE_STRING_H
#include <string.h>
#endif

#include "superwaba.h"

#if defined WIN32
  #include <tchar.h>
#else
  #ifdef UNICODE
  #define _tfopen >>>PGR: FIXME
  #else
  #define _tfopen fopen
  #endif
#endif

#define __DEFINE_FUNCTIONS
#include "SW_messages.h"
#include "SW_palmDb.h"
#include "SW_utils.h"
#include "SW_logging.h"


#if HAVE_GETOPT_LONG

static char usageText[] =
   "Usage: superwaba [OPTION...] [class-name] [class-path] [creator-id]\n\n"
   "\t-w, --width  = <pixels>    window width\n"
   "\t-h, --height = <pixels>    window height\n"
   "\t-x, --xpos = <pixels>      desired window horizontal position\n"
   "\t-y, --ypos = <pixels>      desired window vertical position\n"
   "\t-F, --full                 fullscreen mode\n"
   "\t-p, --path = <dir>         path to the PDB\n"
   "\t-t, --title = <text>       window title\n"
   "\t-v, --version              product version\n"
   "\t--help                     usage help\n";

struct option long_options[] =
{
   {"full",    0, 0, 'F' },
   {"width",   1, 0, 'w' },
   {"height",  1, 0, 'h' },
   {"xpos",    1, 0, 'x' },
   {"ypos",    1, 0, 'y' },
   {"path",    1, 0, 'p' },
   {"pdb",     1, 0, 'n' },
   {"title",   1, 0, 't' },
   {"version", 0, 0, 'v' },
   {"help",    0, 0, 'H' },
   {0,         0, 0, 0   }
};

#else

static char usageText[] =
   "Usage: superwaba [OPTION...] [class-name] [class-path] [creator-id]\n\n"
   "\t/w <pixels>    window width\n"
   "\t/h <pixels>    window height\n"
   "\t/x <pixels>    desired window horizontal position\n"
   "\t/y <pixels>    desired window vertical position\n"
   "\t/F             fullscreen mode\n"
   "\t/p <dir>       path to the PDB\n"
   "\t/t <text>      window title\n"
   "\t/v             product version\n"
   "\t/?             usage help\n";

#endif

char *errorMessages[] =
{
    "Sanity",
    "Incompatible device",
    "Can't access SuperWaba classes",
    "Can't access app classes",
    "Can't allocate memory",
    "Out of class memory",
    "Out of object memory",
    "Native stack overflow",
    "Native stack underflow",
    "Stack overflow",

    "Bad class",
    "Bad opcode",
    "Can't find class",
    "Can't find method",
    "Can't find field",
    "Null object access",
    "Null array access",
    "Class too large",
    "Divide by zero",
    "Class cast exception",

    "Index out of range",
    "Can't allocate object memory",
    "Bad Utf String",
    "No fonts found. Make sure [H,M,L]SW*.pdb font files are located under SuperWaba default directory",
    "Method called on array/Missing class",
    "Can't find native method implementation",
    "Uncaught exception",
    "Class in wrong package",
    "Can't initialize application",
    "No MainWindow object",
    
    "No more memory",
    "SDL init failed",
    "SDLNet init failed",
    "Bad option",
    "No number required",
};

int32 getTimeStamp(); // defined in SW_events
static int recurseDebug = 0; // recurse debug() calls guard

//////////////////////////////////////////////////////////////////////////////
//                             Debug Support                                //
//////////////////////////////////////////////////////////////////////////////

// guich@310_22: now debug stores everything in a txt file in the application's
// directory

// guich superwaba 1.1
#if !defined(PALMOS)         ///////////////// Non PALM OS //////////////////

void createFileName(char *appPath, char *file, TCHAR *path, char *ext);
static void systemOutOpen(uint16 overwrite)
{
   if (vmGlobals->appPath != 0) // fdie@ temporary test to prevent a vmGlobals->appPath when the VM is started with the --help option
   {
      TCHAR tempPath[MAX_PATH];
      createFileName(vmGlobals->appPath, "DebugConsole", tempPath, ".txt");
      vmGlobals->hSystemOut = overwrite?0:_tfopen(tempPath, TEXT("a+")); // guich@501_2: changed rb+ to a+
      if (!vmGlobals->hSystemOut) // if the file does not exists (or we want to overwrite), force its creation
      {
         vmGlobals->hSystemOut = _tfopen(tempPath, TEXT("wb"));
         if (vmGlobals->hSystemOut)
         {
            fclose(vmGlobals->hSystemOut);
            vmGlobals->hSystemOut = _tfopen(tempPath, TEXT("a+")); // we must always keep it open in a+ mode
         }
      }
      vmGlobals->maxMemoLen = 64000;  // unused?
      vmGlobals->stateSystemOut = (vmGlobals->hSystemOut)? SYSOUT_OPENED : SYSOUT_COULDNOTCREATE;
   }
}

void systemOutClose()
{
   if (vmGlobals->stateSystemOut == SYSOUT_OPENED)
   {
      debug("======= End =======\n");
      fclose(vmGlobals->hSystemOut);
      vmGlobals->hSystemOut = 0;
      vmGlobals->stateSystemOut = SYSOUT_CLOSED;
   }
}

void debug(char *msg)   // guich@400
{      
   /* protect before vmGlobals is initialized and after its cleanup and also protect against recurse calls */
   if (vmGlobals == NULL || recurseDebug) return; 
   recurseDebug = 1;

   if (vmGlobals->stateSystemOut == SYSOUT_CLOSED) // start the db
      systemOutOpen(0);
   if ((vmGlobals->stateSystemOut == SYSOUT_OPENED) && msg)
   {
   #if defined(ENABLE_TRACE)
      fwrite(msg, 1, xstrlen(msg), vmGlobals->hSystemOut);
      fflush(vmGlobals->hSystemOut);
   #else
      char buf[10];
      char crlf[] = "\n"; // guich@501_2: a+ mode seems not like \r\n 
      int32 number;

      if (xstrncmp(msg,"!erase debug!",13) == 0) // guich@420_18: delete the debug memo?
      {
         fclose(vmGlobals->hSystemOut);
         systemOutOpen(1);
      }
      else
      {
          // prepend with timestamp + space
          number = vmGlobals->getTimeStamp();
          buf[0] = (char)((number / 100000L  % 10) + '0');
          buf[1] = (char)((number / 10000L   % 10) + '0');
          buf[2] = (char)((number / 1000L    % 10) + '0');
          buf[3] = (char)((number / 100L     % 10) + '0');
          buf[4] = (char)((number / 10L      % 10) + '0');
          buf[5] = (char)((number            % 10) + '0');
          buf[6] = ' ';
          fwrite(buf, 1, 7, vmGlobals->hSystemOut);
          fwrite(msg, 1, xstrlen(msg), vmGlobals->hSystemOut);
          fwrite(crlf, 1, 1, vmGlobals->hSystemOut);
          fflush(vmGlobals->hSystemOut); // guich: make sure this is written immediately
      }
   #endif
   }
   recurseDebug = 0;
}


#else                  /////////////////// PALM OS ////////////////////

static char *titleSystemOut = "SuperWaba Debug Console";
#define IS_TITLE(x) (!xstrncmp((char *)x,titleSystemOut,xstrlen(titleSystemOut)))

static void systemOutOpen()
{
   if (vmGlobals->dmRefSystemOut == 0) // guich@200b4_19
   {
      vmGlobals->dmRefSystemOut = vmGlobals->findAndOpenDB("MemoDB", MEMO_CREATOR_ID, DATA_TYPE_ID, 0, 0);
      vmGlobals->maxMemoLen = (vmGlobals->romVersion >= 0x03000000) ? 32000 : 4000;
   }

   if (vmGlobals->dmRefSystemOut)
   {
      // finds the position to insert our data
      VoidHand recH;
      VoidPtr recPtr;
      int i;
      int n = DmNumRecords(vmGlobals->dmRefSystemOut);

      vmGlobals->stateSystemOut = SYSOUT_NOTFOUND;
      for (i =0; i < n; i++)
      {
         recH = DmGetRecord(vmGlobals->dmRefSystemOut, i);
         //DbgMessage((*vmGlobals->int2str)("systemOutOpen handle:",(uint32)recH));
         if (recH) // not deleted?
         {
            recPtr = MemHandleLock(recH);
            if (!recPtr)
               DmReleaseRecord(vmGlobals->dmRefSystemOut, i, false);
            else
            if (IS_TITLE(recPtr)) // "waba.system.out\n".length
            {
               vmGlobals->posSystemOut = i;
               MemHandleUnlock(recH);
               DmReleaseRecord(vmGlobals->dmRefSystemOut, vmGlobals->posSystemOut, false);
               vmGlobals->stateSystemOut = SYSOUT_OPENED;
               break;
            }
            else
            {
               MemHandleUnlock(recH);
               DmReleaseRecord(vmGlobals->dmRefSystemOut, i, false);
            }
         }
      }

      if (vmGlobals->stateSystemOut == SYSOUT_NOTFOUND)
      {
         vmGlobals->posSystemOut = n;
         recH = DmNewRecord(vmGlobals->dmRefSystemOut, &vmGlobals->posSystemOut, 1);
         if (!recH)
            vmGlobals->stateSystemOut = SYSOUT_COULDNOTCREATE;
         else
         {
            recPtr = MemHandleLock(recH);
            if (!recPtr)
            {
               DmDeleteRecord(vmGlobals->dmRefSystemOut, vmGlobals->posSystemOut);
               vmGlobals->stateSystemOut = SYSOUT_COULDNOTCREATE;
            }
            else
            {
               int recurseSave;
               // everything is ok. release the record.
               MemHandleUnlock(recH);
               DmReleaseRecord(vmGlobals->dmRefSystemOut, vmGlobals->posSystemOut, false);
               vmGlobals->stateSystemOut = SYSOUT_OPENED;
               recurseSave = recurseDebug; // save, then disable recurse lock
               recurseDebug = 0;
               // write the header
               debug(titleSystemOut);
               recurseDebug = recurseSave;
            }
         }
      }
      if (vmGlobals->stateSystemOut != SYSOUT_OPENED)
         DmCloseDatabase(vmGlobals->dmRefSystemOut); // close the database
   }
}

void systemOutClose()
{
   if (vmGlobals->stateSystemOut == SYSOUT_OPENED || vmGlobals->stateSystemOut == SYSOUT_ERASED || vmGlobals->dmRefSystemOut) // guich@200b4_19
   {
      if (vmGlobals->stateSystemOut == SYSOUT_OPENED)
         debug("======= End =======\n");
      DmCloseDatabase(vmGlobals->dmRefSystemOut);
      vmGlobals->dmRefSystemOut = NULL;
      vmGlobals->stateSystemOut = SYSOUT_CLOSED;
   }
}

void debug(char *msg)   // guich@310_22
{
   static char buf[10];
   char *cr0 = "\n\0";
   int slen = msg?xstrlen(msg):0;

   /* protect before vmGlobals is initialized and after its cleanup and also protect against recurse calls */
   if (vmGlobals == NULL || vmGlobals->findAndOpenDB == NULL || recurseDebug) return; 
   recurseDebug = 1;

   //_LOG_INFO((LOG_CAT_MSG, msg));
   #ifdef ENABLE_TRACE
   {
      static int count=10;
      DbgMessage(msg);
      DbgMessage("\n");
      if (count-- == 0)
      {
         SysTaskDelay(10);
         count = 10;
      }
   }
   #else
   if (vmGlobals->stateSystemOut == SYSOUT_CLOSED) // start the db
   {
      systemOutOpen();
   }
   if (vmGlobals->stateSystemOut == SYSOUT_OPENED)
   {
      VoidHand recH;
      VoidPtr recPtr;

      if (xstrncmp(msg,"!erase debug!",13) == 0) // guich@420_18: delete the debug memo?
      {
         DmRemoveRecord(vmGlobals->dmRefSystemOut, vmGlobals->posSystemOut);
         vmGlobals->stateSystemOut = SYSOUT_ERASED;
         systemOutClose();
      }
      else
      {
          recH = DmGetRecord(vmGlobals->dmRefSystemOut, vmGlobals->posSystemOut);
          if (recH) // not deleted?
          {
             recPtr = MemHandleLock(recH);
             if (!recPtr)
                DmReleaseRecord(vmGlobals->dmRefSystemOut, vmGlobals->posSystemOut, false);
             else
             {
                ULong totalLen = slen+2;  // plus 2 for the cr0
                ULong lenH = MemHandleSize(recH)-1; // minus 1 for the \0
                int isTitle = IS_TITLE(msg);
                totalLen += lenH;
                if (totalLen > vmGlobals->maxMemoLen) // guich@200b2: dont let the memo grow too much. palm os 2.0 hangs!
                {
                   lenH = xstrlen(titleSystemOut)+2; // keep title
                   totalLen = slen+2+lenH;
                }
    
                if (!isTitle) totalLen += 7; // size of timestamp
                MemHandleUnlock(recH);
                DmReleaseRecord(vmGlobals->dmRefSystemOut, vmGlobals->posSystemOut, false);
                recH = DmResizeRecord(vmGlobals->dmRefSystemOut, vmGlobals->posSystemOut, totalLen); // open space to the new string
                if (recH)
                {
                   recH = DmGetRecord(vmGlobals->dmRefSystemOut, vmGlobals->posSystemOut); // guich@300_46
                   if (recH)
                   {
                      recPtr = MemHandleLock(recH);
                      if (!recPtr)
                         DmReleaseRecord(vmGlobals->dmRefSystemOut, vmGlobals->posSystemOut, false);
                      else
                      {
                         // prepend with timestamp + space
                         if (!isTitle)
                         {
                            int32 number = getTimeStamp();
                            buf[7] = 0;
                            buf[6] = ' ';
                            buf[5] = (number            % 10) + '0';
                            buf[4] = (number / 10L      % 10) + '0';
                            buf[3] = (number / 100L     % 10) + '0';
                            buf[2] = (number / 1000L    % 10) + '0';
                            buf[1] = (number / 10000L   % 10) + '0';
                            buf[0] = (number / 100000L  % 10) + '0';
                            DmWrite(recPtr, lenH, buf, 7); lenH += 7;
                         }
                         DmWrite(recPtr, lenH, msg, slen); lenH += slen;
                         DmWrite(recPtr, lenH, cr0, 2);
                         MemHandleUnlock(recH);
                         DmReleaseRecord(vmGlobals->dmRefSystemOut, vmGlobals->posSystemOut, false);
                      }
                   }
                }
             }
          }
      }
   }
   #endif // ENABLE_TRACE
   recurseDebug = 0;
}

#endif


//////////////////////////////////////////////////////////////////////////////
//                                 Error Messages                           //
//////////////////////////////////////////////////////////////////////////////
void sendError(ErrorStatus * status)
{
   static int sent=0;
   char buf[500];  //>>PGR: was 384.  Not sure if it's enough.
   char * max = buf + sizeof buf;
   char * cur;
   char const * msg;

   if (sent) return; // avoid duplicate
   cur = boundedCopy(buf, max, "An error was found in the program being\n"
                               "run by the SuperWaba VM.\n\nError: ");
   msg = errorMessages[status->errNum - 1];
   if (msg)
      cur = boundedCat(cur, max, msg);
   msg = status->arg1;
   if (msg[0])
   {
      cur = boundedCat(cur, max, " ");
      cur = boundedCat(cur, max, msg);
   }
   msg = status->arg2;
   if (msg[0])
   {
      cur = boundedCat(cur, max, "\n");
      cur = boundedCat(cur, max, msg);
      cur = boundedCat(cur, max, "\n");
   }
   if (status->className[0] || status->methodName[0])
   {
      cur = boundedCat(cur, max, "in\n");
      msg = status->className;
      if (msg[0])
         cur = boundedCat(cur, max, msg);
      msg = status->methodName;
      if (msg[0])
      {
         cur = boundedCat(cur, max, ".");
         cur = boundedCat(cur, max, msg);
      }
   }
   if (status->errNum == ERR_CantAccessCoreClasses)
      msg = "The SuperWaba VM is not fully installed.\n"
            "Please make sure the file SuperWaba.pdb\n"
            "is installed under '\\SuperWaba\n";
   else
   if (status->errNum == ERR_UncaughtException &&
      !xstrncmp(status->arg2,"Can't find class",16)) // guich@350_31: give a better error message. note: the error can be a Uncaught exception
   {
      msg = "Check if the application's .PDB\n"
            "and/or a required library (E.G.\n"
            "XPlat*.pdb or other)\n"
            "are correctly installed.\n\n"
            "See debug console for details\n(a txt in the program's folder)\n";
   }
   else
   {
      msg = "\n\nPlease notify the program's author.\n"
            "See debug console for details\n(a txt in the program's folder)\n";
   }
   cur = boundedCat(cur, max, msg);
   debug(buf);
// postStopEvent(status->errNum);
   sent = 1;
}

/*--------------------------------------------------------------------version-+
|                                                                             |
+----------------------------------------------------------------------------*/
void version()
{
   char buf[500];
   char * max = buf + sizeof buf;
   char * cur;

   cur = boundedCopy(buf, max, "Superwaba VM version ");

   // define VERSION for non GNU systems until we have a platform specific
   // predefined "config.h.platform"
   #ifndef HAVE_CONFIG_H
   #define VERSION "5.00"
   #endif
   cur = boundedCat(cur, max, VERSION);
   cur = boundedCat(cur, max, " (c) 2004 \t=> www.superwaba.com\n");
   debug(buf);
#ifdef linux
   printf(buf);
#endif
}

/*----------------------------------------------------------------------usage-+
|                                                                             |
+----------------------------------------------------------------------------*/
void usage()
{
   version();
   debug(usageText);
#ifdef linux
   printf(usageText);
#endif
}

/*-----------------------------------------------------------------showOption-+
|                                                                             |
+----------------------------------------------------------------------------*/
#if HAVE_GETOPT_LONG
void showOption(char const * option, char const * optarg)
{
   char buf[200];
   char * max = buf + sizeof buf;
   char * cur;

   cur = boundedCopy(buf, max, "option ");
   cur = boundedCat(cur, max, option);
   if (optarg)
   {
      cur = boundedCat(cur, max, " with arg ");
      cur = boundedCat(cur, max, optarg);
   }
   cur = boundedCat(cur, max, "\n");
   debug(buf);
#ifdef linux
   printf(buf);
#endif
}
#endif

/*----------------------------------------------------------------------trace-+
|                                                                             |
+----------------------------------------------------------------------------*/
#ifdef ENABLE_TRACE
void trace(WClass *iclass, WClassMethod *imethod)
{
   int j=0,l;
   char *c;
   uint16 idx;
//#ifdef PALMOS
   static char traceBuf[200];
/*#else
   uint16 traceBuf[200];
#endif*/
   UtfString methodName = (imethod==NULL)?(*vmGlobals->createUtfString)("new "):(*vmGlobals->getUtfString)(iclass, (uint16)((idx=METH_nameIndex(imethod)) < iclass->numConstants ? idx : 0));
   UtfString methodDesc = (imethod==NULL)?(*vmGlobals->createUtfString)("")    :(*vmGlobals->getUtfString)(iclass, (uint16)((idx=METH_descIndex(imethod)) < iclass->numConstants ? idx : 0));

   // finds the last '/' of the class. ignores before it
   c = iclass->className.str;
   while (*c && *c != '/') c++;
   if (*c == '/') // there exist a '/'?
   {
      l = iclass->className.len;
      c = &iclass->className.str[l-1];
      while (*c != '/' && l > 0) {l--; c--;}
      if (*c == '/') c++;
   }
   // concats the strings
   for (; *c; j++)
      traceBuf[j] = *c++;
   j--;
   traceBuf[j++] = '.';
   c = methodName.str;
   for (; *c && (j < 99); j++,c++)
      traceBuf[j] = *c;
   j--;
   c = methodDesc.str;
   if (*c)
   {
      for (; *c && (j < 99); j++)
         traceBuf[j] = *c++;
      c--; j--;
   }
   traceBuf[j++] = '\r';
   traceBuf[j++] = '\n';
   traceBuf[j] = 0;
   debug(traceBuf);
//   if (xstrncmp(traceBuf,"Rect.<init>(IIII)V",18) == 0) // @ t
//      *traceBuf = 0;
}
#endif

/*-----------------------------------------------------------------printToBuf-+
| Description:                                                                |
|   copies one or two UtfStrings (by concatenation) into a character          |
|   buffer (null terminated) suitable for output                              |
+----------------------------------------------------------------------------*/
void printToBuf(char *buf, int32 maxbuf, UtfString *s1, UtfString *s2)
{
    int32 len, len2;

    len = 0;
    if (s1 != NULL && s1->str != NULL && s1->len > 0)
    {
       len = s1->len;
       if (len >= maxbuf)
          len = maxbuf - 1;
       xstrncpy(buf, s1->str, len);
    }
    if (s2 != NULL && s2->str != NULL && s2->len > 0)
    {
       len2 = s2->len;
       if (len2 + len >= maxbuf)
          len2 = maxbuf - len - 1;
       xstrncpy(&buf[len], s2->str, len2);
       len += len2;
    }
    buf[len] = 0;
}
/*===========================================================================*/
