/*********************************************************************************
 *  SuperWaba Virtual Machine, version 5                                         *
 *  Copyright (C) 2000-2005 SuperWaba Ltda.                                      *
 *  All Rights Reserved                                                          *
 *                                                                               *
 *  This library and virtual machine is distributed in the hope that it will     *
 *  be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                         *
 *                                                                               *
 *  Please see the software license located at /license.txt for more details.    *
 *                                                                               *
 *  You should have received a copy of the License along with this software;     *
 *  if not, write to                                                             *
 *                                                                               *
 *     SuperWaba Ltda.                                                           *
 *     Rua Padre Leonel Franca, 480, sala 27A                                    *
 *     Instituto Genesis - PUC-RIO                                               *
 *     Rio de Janeiro / RJ - Brazil                                              *
 *     Cep: 22451-000                                                            *
 *     E-mail: licensing@superwaba.com.br                                        *
 *********************************************************************************/

#ifndef SW_PALMDB_H
#define SW_PALMDB_H

/*

This is a Palm Database (PDB) support for SuperWaba WinCE native libraries.
You will call any of the functions as if they were local, but the macros
will redirect the call to the vmGlobals structure. This will maintain compatibility
with existing PalmOS code.

The declarations must be customized based on some defines:

Documentation:

	HandlePDBSearchProcType: callback function for DmListDatabases. Return: 0 to continue the search, 1 to stop it
	DmCloseDatabase:         closes the given database
	DmCreateDatabase:        create a database with the given parameters. Name must have 32 characters (counting \0)
	DmOpenDatabase:          opens the given database. It must have the full pathname specified
	DmOpenDatabaseReadOnly:  opens the given database for READ ONLY. It must have the full pathname specified
	DmReleaseRecord:         release the record. If the database is read-only, it can have many
	                         locked records. Otherwise, if it is read-write, it can have only one record
	DmNewRecord:             creates a new record at the given position. if *atP is < 0, the record is appended
	DmRenameDatabase:        Rename the given opened database to the new parameters
	DmDeleteDatabase:        delete the given opened database
	DmListDatabases:         lists the databases, calling the defined proc. If proc is null,
                            only the total files is returned. userVars can be any pointer
                            to a user defined variable. if recursive is 1, pdb files are
                            searched from SUPERWABA_CORE_PATH recursively
	DmListDatabasesIn:       lists the databases, calling the defined proc. If proc is null,
                            only the total files is returned. userVars can be any pointer
                            to a user defined variable. if recursive is 1, pdb files are
                            searched from the given path recursively
	DmFindAndOpenDB:         Finds and opens the database with the given parameters
	DmNumRecords:            returns the total number of records in the database
	DmWrite:                 writes to the given opened database
	DmResizeRecord:          resizes the given record. The record cannot be locked
	DmRemoveRecord:          removes the given record
	DmGetRecord:             gets a handle to the given record. The record is marked as locked
                            and is copied into a buffer. You must call DmReleaseRecord to
                            unlock this record and also you must free the record by calling xfree
   DmQueryRecord:           gets a handle to the given record. The record may be locked. size ptr
                            may be NULL, or a variable set to 0 (to load the entire record) or the number of
                            bytes to load. Note that you must free the allocated MemHandle by yourself, using xfree
	DmGetRecordAttributes:   get the record attributes, even if the record is locked
	DmSetRecordAttributes:   set the record attributes, even if the record is locked
	DmSetAttributes:         set the database attributes
	DmGetAttributes:         get the database attributes
	DmSetBackupAttribute:    sets the backup attribute of the database
	DmGetRecordSize:         returns the record size of the given record
	DmGetDatabaseTypeCreatorCount: returns the database type and creator for the given filename.
                            Optionally, returns also the records count
	DmGetName:               returns the internal name of the file. The buffer must be 32 chars long
	splitPath:               splits the path from the filename. path or name may be null

*/

// In Windows, != 0 means success, in PalmOS 0 means success
// This defines a standard approach
#define PALM_SUCCESS 0
#define PALM_ERROR 1

#ifndef PALMOS
 #include <stdio.h>
#endif

typedef byte (*HandlePDBSearchProcType) (TCHAR *fName, void *userVars);
void splitPath(TCHAR *fullPath, TCHAR *path, TCHAR *name);

#ifndef PALMOS

 typedef struct t_Catalog* DmOpenRef;

 #define	dmRecAttrDelete			0x80	// delete this record next sync
 #define	dmRecAttrDirty				0x40	// archive this record next sync
 #define	dmRecAttrBusy				0x20	// record currently in use
 #define	dmRecAttrSecret			0x10	// "secret" record - password protected

 #define MemHandleLock(x) x
 #define MemHandleUnlock(x)

 #define DmCloseDatabase(db) VM_CALL(_DmCloseDatabase)(db)
 #define DmCreateDatabase(name, creator, type) VM_CALL(_DmCreateDatabase)(name, creator, type)
 #define DmDeleteDatabase(db) VM_CALL(_DmDeleteDatabase)(db)
 #define DmFindAndOpenDB(name, creator, type, readOnly, onCard) VM_CALL(findAndOpenDB)(name, creator, type, readOnly, onCard)
 #define DmGetAttributes(db, attrs) VM_CALL(_DmGetAttributes)(db, attrs)
 #define DmGetDatabaseTypeCreatorCount(fullPath, type, creator, recCount) VM_CALL(_DmGetDatabaseTypeCreatorCount)(fullPath, type, creator, recCount)
 #define DmGetName(db, buf) VM_CALL(_DmGetName)(db, buf)
 #define DmGetRecord(db, index) VM_CALL(_DmGetRecord)(db, index)
 #define DmGetRecordAttributes(db, index, attrs) VM_CALL(_DmGetRecordAttributes)(db, index, attrs)
 #define DmGetRecordSize(db, index) VM_CALL(_DmGetRecordSize)(db, index)
 #define DmListDatabases(proc, userVars, recursive) VM_CALL(_DmListDatabases)(proc, userVars, recursive)
 #define DmListDatabasesIn(path, proc, userVars, recursive) VM_CALL(_DmListDatabasesIn)(path, proc, userVars, recursive)
 #define DmNewRecord(db, atP, size) VM_CALL(_DmNewRecord)(db, atP, size)
 #define DmNumRecords(db) VM_CALL(_DmNumRecords)(db)
 #define DmOpenDatabase(fileName, creator, type) VM_CALL(_DmOpenDatabase)(fileName, creator, type)
 #define DmOpenDatabaseReadOnly(fileName, creator, type) VM_CALL(_DmOpenDatabaseReadOnly)(fileName, creator, type)
 #define DmQueryRecord(db, index, size) VM_CALL(_DmQueryRecord)(db, index, size)
 #define DmQueryRecordPart(db, index, size, ofs, buf) VM_CALL(_DmQueryRecordPart)(db, index, size, ofs, buf)
 #define DmReleaseRecord(db, index, dirty) VM_CALL(_DmReleaseRecord)(db, index, dirty)
 #define DmRemoveRecord(db, index) VM_CALL(_DmRemoveRecord)(db, index)
 #define DmRenameDatabase(db, newType, newCreator, newName) VM_CALL(_DmRenameDatabase)(db, newType, newCreator, newName)
 #define DmResizeRecord(db, index, size) VM_CALL(_DmResizeRecord)(db, index, size)
 #define DmSetAttributes(db, attrs) VM_CALL(_DmSetAttributes)(db, attrs)
 #define DmSetBackupAttribute(db) VM_CALL(_DmSetBackupAttribute)(db)
 #define DmSetRecordAttributes(db, index, attrs) VM_CALL(_DmSetRecordAttributes)(db, index, attrs)
 #define DmWrite(db, index, offset, srcP, bytes) VM_CALL(_DmWrite)(db, index, offset, srcP, bytes)
 #define splitPath(fullPath, path, name) VM_CALL(_splitPath)(fullPath, path, name)

#endif

typedef DmOpenRef (*DmFindAndOpenDBProcType)(char *name, uint32 creator, uint32 type, byte readOnly, byte onCard);
typedef char* (*DmGetNameProcType)(DmOpenRef db, char *buf);
typedef DmOpenRef (*DmOpenDatabaseProcType)(TCHAR *fileName, uint32 creator, uint32 type);
typedef DmOpenRef (*DmOpenDatabaseReadOnlyProcType)(TCHAR *fileName, uint32 creator, uint32 type);
typedef Err (*DmCloseDatabaseProcType)(DmOpenRef db);
typedef Err (*DmCreateDatabaseProcType)(char *name, uint32 creator, uint32 type);
typedef Err (*DmDeleteDatabaseProcType)(DmOpenRef db);
typedef Err (*DmGetAttributesProcType)(DmOpenRef db, int16 *attrs);
typedef Err (*DmGetDatabaseTypeCreatorCountProcType)(TCHAR *fullPath, uint32 *type, uint32 *creator, uint16 *recCount);
typedef Err (*DmGetRecordAttributesProcType)(DmOpenRef db, uint16 index, byte *attrs);
typedef Err (*DmReleaseRecordProcType)(DmOpenRef db, int32 index, byte dirty);
typedef Err (*DmRemoveRecordProcType)(DmOpenRef db, uint16 index);
typedef Err (*DmRenameDatabaseProcType)(DmOpenRef db, uint32 newType, uint32 newCreator, char *newName);
typedef Err (*DmSetAttributesProcType)(DmOpenRef db, int16 attrs);
typedef Err (*DmSetBackupAttributeProcType)(DmOpenRef db);
typedef Err (*DmSetRecordAttributesProcType)(DmOpenRef db, uint16 index, byte attrs);
typedef Err (*DmWriteProcType)(DmOpenRef db, uint16 index, uint32 offset, const void *srcP, uint32 bytes);
typedef int32 (*DmGetRecordSizeProcType)(DmOpenRef db, uint16 index);
typedef int32 (*DmListDatabasesInProcType)(char *path, HandlePDBSearchProcType proc, void *userVars, byte recursive);
typedef int32 (*DmListDatabasesProcType)(HandlePDBSearchProcType proc, void *userVars, byte recursive);
typedef MemHandle (*DmGetRecordProcType)(DmOpenRef db, uint16 index);
typedef MemHandle (*DmNewRecordProcType)(DmOpenRef db, uint16 *atP, uint32 size);
typedef MemHandle (*DmQueryRecordProcType)(DmOpenRef db, uint16 index, int32 *size);
typedef Err (*DmQueryRecordPartProcType)(DmOpenRef db, uint16 index, int32 size, int32 ofs, byte* buf);
typedef MemHandle (*DmResizeRecordProcType)(DmOpenRef db, uint16 index, uint32 size);
typedef uint16 (*DmNumRecordsProcType)(DmOpenRef db);
typedef void (*splitPathProcType)(TCHAR *fullPath, TCHAR *path, TCHAR *name);

#endif
