#ifndef ENDIANESS_H
#define ENDIANESS_H

/*
| Endiannesses
|
| On ARM, while each word of a double is stored little-endian, the most
| significant word (the word containing the sign bit) is at the lower address.
| This is neither pure big-endian (as implemented in 68k processors) nor pure
| little-endian (as implemented in x86 processors).
| Litterature describes it as "mix" or "cross" endianess.
|
| The MARM layout comes from the design of the ARM floating point coprocessor,
| which specifies that doubles are stored in memory as two 32-bit words, with
| the first word containing the top 32 bits, the second word the low 32 bits.
| EPOC uses little-endian storage for words, giving rise to the rather strange
| mixture of big-endian and little-endian layout.
*/

/* The two types of endianness */
#define SW_LIL_ENDIAN	1234
#define SW_BIG_ENDIAN	4321

/* Pardon the mess, I'm trying to determine the endianness of this host.
   I'm doing it by preprocessor defines rather than some sort of configure
   script so that application code can use this too.  The "right" way would
   be to dynamically generate this file on install, but that's a lot of work.
 */
#if  defined(__i386__) || defined(__ia64__) || defined(WIN32) || \
    (defined(__alpha__) || defined(__alpha)) || \
     defined(__arm__) || \
    (defined(__mips__) && defined(__MIPSEL__)) || \
     defined(__SYMBIAN32__) || \
     defined(__x86_64__) || \
     defined(__LITTLE_ENDIAN__)
#define SW_BYTEORDER	SW_LIL_ENDIAN
#else
#define SW_BYTEORDER	SW_BIG_ENDIAN
#endif

#if SW_BYTEORDER == SW_LIL_ENDIAN && !defined __SYMBIAN32__
  // && !defined __arm__   ???
  #define SW_WORDORDER SW_LIL_ENDIAN
#else  // BIG endian machines (68K), or Mixed (ARM)
  #define SW_WORDORDER SW_BIG_ENDIAN
#endif

// align same byte endianness
static __inline uint16 sw_align16(byte * b)
{
   uint16 tgt;
   *((byte *)&tgt+0) = b[0];
   *((byte *)&tgt+1) = b[1];
   return tgt;
}

// align diff byte endianness
static __inline uint16 sw_align16_r(byte * b)
{
   uint16 tgt;
   *((byte *)&tgt+0) = b[1];
   *((byte *)&tgt+1) = b[0];
   return tgt;
}

// align same byte endianness
static __inline uint32 sw_align32(byte * b)
{
   uint32 tgt;
   *((byte *)&tgt+0) = b[0];
   *((byte *)&tgt+1) = b[1];
   *((byte *)&tgt+2) = b[2];
   *((byte *)&tgt+3) = b[3];
   return tgt;
}

// align diff byte endianness
static __inline uint32 sw_align32_r(byte * b)
{
   uint32 tgt;
   *((byte *)&tgt+0) = b[3];
   *((byte *)&tgt+1) = b[2];
   *((byte *)&tgt+2) = b[1];
   *((byte *)&tgt+3) = b[0];
   return tgt;
}

// align same byte endianness, same word-endianness (pure endiannesses)
static __inline int64 sw_align64(byte * b)
{
   int64 tgt;
   *((byte *)&tgt+0) = b[0];
   *((byte *)&tgt+1) = b[1];
   *((byte *)&tgt+2) = b[2];
   *((byte *)&tgt+3) = b[3];
   *((byte *)&tgt+4) = b[4];
   *((byte *)&tgt+5) = b[5];
   *((byte *)&tgt+6) = b[6];
   *((byte *)&tgt+7) = b[7];
   return tgt;
}

// align diff byte endianness, diff word-endianness (pure endiannesses)
static __inline int64 sw_align64_r(byte * b)
{
   int64 tgt;
   *((byte *)&tgt+0) = b[7];
   *((byte *)&tgt+1) = b[6];
   *((byte *)&tgt+2) = b[5];
   *((byte *)&tgt+3) = b[4];
   *((byte *)&tgt+4) = b[3];
   *((byte *)&tgt+5) = b[2];
   *((byte *)&tgt+6) = b[1];
   *((byte *)&tgt+7) = b[0];
   return tgt;
}

// align same byte endianness, diff word-endianness (mix endiannesses)
static __inline int64 sw_align64_w(byte * b)
{
   int64 tgt;
   *((byte *)&tgt+0) = b[4];
   *((byte *)&tgt+1) = b[5];
   *((byte *)&tgt+2) = b[6];
   *((byte *)&tgt+3) = b[7];
   *((byte *)&tgt+4) = b[0];
   *((byte *)&tgt+5) = b[1];
   *((byte *)&tgt+6) = b[2];
   *((byte *)&tgt+7) = b[3];
   return tgt;
}

// align diff byte endianness, same word-endianness (mix endiannesses)
static __inline int64 sw_align64_b(byte * b)
{
   int64 tgt;
   *((byte *)&tgt+0) = b[3];
   *((byte *)&tgt+1) = b[2];
   *((byte *)&tgt+2) = b[1];
   *((byte *)&tgt+3) = b[0];
   *((byte *)&tgt+4) = b[7];
   *((byte *)&tgt+5) = b[6];
   *((byte *)&tgt+6) = b[5];
   *((byte *)&tgt+7) = b[4];
   return tgt;
}

/*
| Read a numeric with the specified endianness
*/
#if SW_BYTEORDER == SW_LIL_ENDIAN
  #define READ_LE16(b)       sw_align16((byte*)b)
  #define READ_BE16(b)       sw_align16_r((byte*)b)
  #define READ_LE32(b)       sw_align32((byte*)b)
  #define READ_BE32(b)       sw_align32_r((byte*)b)
  #if SW_WORDORDER == SW_LIL_ENDIAN
    #define READ_LE64(b)     sw_align64((byte*)b)
    #define READ_LEMix64(b)  sw_align64_w((byte*)b)
    #define READ_BE64(b)     sw_align64_r((byte*)b)
    #define READ_BEMix64(b)  sw_align64_b((byte*)b)
  #else
    #define READ_LE64(b)     sw_align64_w((byte*)b)
    #define READ_LEMix64(b)  sw_align64((byte*)b)
    #define READ_BE64(b)     sw_align64_b((byte*)b)
    #define READ_BEMix64(b)  sw_align64_r((byte*)b)
  #endif
#else
  #define READ_BE16(b)       sw_align16((byte*)b)
  #define READ_LE16(b)       sw_align16_r((byte*)b)
  #define READ_BE32(b)       sw_align32((byte*)b)
  #define READ_LE32(b)       sw_align32_r((byte*)b)
  #if SW_WORDORDER != SW_LIL_ENDIAN
    #define READ_BE64(b)     sw_align64((byte*)b)
    #define READ_BEMix64(b)  sw_align64_w((byte*)b)
    #define READ_LE64(b)     sw_align64_r((byte*)b)
    #define READ_LEMix64(b)  sw_align64_b((byte*)b)
  #else
    #define READ_BE64(b)     sw_align64_w((byte*)b)
    #define READ_BEMix64(b)  sw_align64((byte*)b)
    #define READ_LE64(b)     sw_align64_b((byte*)b)
    #define READ_LEMix64(b)  sw_align64_r((byte*)b)
  #endif
#endif

#define READ_LE_UINT16(b)              READ_LE16((byte*)b)
#define READ_BE_UINT16(b)              READ_BE16((byte*)b)
#define READ_LE_UINT32(b)              READ_LE32((byte*)b)
#define READ_BE_UINT32(b)              READ_BE32((byte*)b)
#define READ_LE_UINT64(b)      (uint64)READ_LE64((byte*)b)
#define READ_LEMix_UINT64(b)   (uint64)READ_LEMix64((byte*)b)
#define READ_BE_UINT64(b)      (uint64)READ_BE64((byte*)b)
#define READ_BEMix_UINT64(b)   (uint64)READ_BEMix64((byte*)b)

#define READ_LE_INT16(b)        (int16)READ_LE16((byte*)b)
#define READ_BE_INT16(b)        (int16)READ_BE16((byte*)b)
#define READ_LE_INT32(b)        (int32)READ_LE32((byte*)b)
#define READ_BE_INT32(b)        (int32)READ_BE32((byte*)b)
#define READ_LE_INT64(b)               READ_LE64((byte*)b)
#define READ_LEMix_INT64(b)            READ_LEMix64((byte*)b)
#define READ_BE_INT64(b)               READ_BE64((byte*)b)
#define READ_BEMix_INT64(b)            READ_BEMix64((byte*)b)

#endif