/*********************************************************************************
 *  SuperWaba Virtual Machine, version 5                                         *
 *  Copyright (C) 2000-2005 SuperWaba Ltda.                                      *
 *  All Rights Reserved                                                          *
 *                                                                               *
 *  This library and virtual machine is distributed in the hope that it will     *
 *  be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                         *
 *                                                                               *
 *  Please see the software license located at /license.txt for more details.    *
 *                                                                               *
 *  You should have received a copy of the License along with this software;     *
 *  if not, write to                                                             *
 *                                                                               *
 *     SuperWaba Ltda.                                                           *
 *     Rua Padre Leonel Franca, 480, sala 27A                                    *
 *     Instituto Genesis - PUC-RIO                                               *
 *     Rio de Janeiro / RJ - Brazil                                              *
 *     Cep: 22451-000                                                            *
 *     E-mail: licensing@superwaba.com.br                                        *
 *********************************************************************************/
/* $Id: SW_Font.c,v 1.8 2005/08/21 20:09:38 fdie Exp $ */

#include "superwaba.h"
#include "SW_natives.h"
#include "SW_palmDb.h"
#include "SW_logging.h"
#include "SDL.h"
#include "SDL_endian.h"

#if HAVE_CONFIG_H
#include "config.h"
#endif

#if HAVE_STRING_H
#include <string.h>
#endif

#if defined PALMOS //----------------------------------------------------------------------

//guich: this must now be selected at runtime
//#if !defined(USE_INTERNAL_FONT)
//#define USE_INTERNAL_FONT 1	// use internal highres fonts by default (0=disabled, 1=lowres fonts, 2=highres fonts)
//#endif

///////////////////////////////////////////////////////////
//          New Unicode Support

// Load and cache used ranges. pending: unload unused ranges
// Return index in range arrays, or -1 if MCS
static int32 uffEnsureRange(char16 ch, struct UserFont *uf)
{
   struct UffUserFont *uff = uf->uff;
   // check if the range is already in the cache - if so, it is sure it is already loaded.
   if (ch <= uff->recentLastChar && ch >= uff->recentFirstChar)
   {
      return uff->recentRangeIndex;
   }
   else
   {
      int32 rindex = -1;

      // Find range. Perform a binary search here. This whould help if a lot of ranges.
      int32 low = 0;
      int32 hi = uff->nParts - 1;
      int32 mid = 0;

      while (low <= hi)
      {
         mid = (low + hi) >> 1;
         if (ch >= uff->uffFirstChar[mid] && ch <= uff->uffLastChar[mid])
         {
            rindex = mid;
            break;
         }
         else
         if (ch < uff->uffFirstChar[mid])
         {
            hi = mid - 1;
         }
         else
         {
            low = mid + 1;
         }
      }

      if (rindex != -1)
      {
         if (uff->ranges[rindex].rowWIB == 0) // not yet loaded
         {
            int32 size;
            struct UffUserFont *uff = uf->uff;
            struct UserFontRange *range = &uff->ranges[rindex];
            byte *recPtr;
            int32 *bit;
            range->recH = DmQueryRecord(uf->dmRef, uff->recNo[rindex]); // load record
            recPtr = (byte*)MemHandleLock(range->recH);
            range->rowWIB = READ_BE_UINT32(recPtr); recPtr += 4;

            size = (int32)range->rowWIB * (int32)uf->fontP->fRectHeight;
            range->bitmapTable = recPtr; recPtr += size;

            // unfortunately, we can't point to the bitIndexTable due to memory alinment
            size = uff->uffLastChar[rindex] - uff->uffFirstChar[rindex] + 1 + 1;
            range->bitIndexTable = bit = (int32*)vmGlobals->xmalloc(size<<2);
            if (bit == NULL)
            {
               MemHandleUnlock(range->recH);
               return -1;
            }
            for (; size--; recPtr+=4)
            {
               *bit++ = READ_BE_INT32(recPtr);
            }
         }
         uff->recentFirstChar  = uff->uffFirstChar[rindex];
         uff->recentLastChar   = uff->uffLastChar[rindex];
         uff->recentRangeIndex = rindex;
      }
      return rindex;
   }
}

// Get the width of a character
static int uffCharWidth(char16 ch, struct UserFont *uf)
{
   struct UffUserFont *uff = uf->uff;
   // ensure that range is loaded
   int32 rindex = uffEnsureRange(ch,uf);
   if (rindex == -1)
   {
      return UF_MCS_WIDTH;
   }
   else
   {
      struct UserFontRange *range = &uff->ranges[rindex];
      int32 index = (int32)ch - uff->recentFirstChar;
      return range->bitIndexTable[index + 1] - range->bitIndexTable[index];
   }
}

// Update a CharBits object to reflect the specificities of a given character.
void uffSetCharBits(char16 ch, struct UserFont *uf, struct CharBits *bits)
{
   struct UffUserFont *uff = uf->uff;
   int32 rindex = uffEnsureRange(ch,uf);
   if (rindex == -1 && ch != 0xFFFD) // eisvogel@421_44: The character could not be loaded by the ensureRange function. In Unicode there are several replacement characters. The interesting one is U+FFFD. If a font can't display a requested character it should display U+FFFD.
   {
   	ch = 0xFFFd; // unicode replacement character
     	rindex = uffEnsureRange(ch,uf);
	}
	if (rindex == -1 && ch != 0x0) // eisvogel@421_44: Neither the requested character nor the replacement character was found in the font. Usuall TTF and BDF fonts define a so called ".nodef" character at the codepoint 0x0. This is the very last fallback for characters not supported by the font.
   {
   	ch=0x0; // ".notdef" character in ttf and bdf fonts
     	rindex = uffEnsureRange(ch,uf);
	}
   if (rindex == -1)
   {
      bits->width = 0;
   }
   else
   {
      int32 cindex = (int32) ch - uff->recentFirstChar;
      struct UserFontRange *range = &uff->ranges[rindex];
      bits->offset = range->bitIndexTable[cindex];
      bits->width = range->bitIndexTable[cindex + 1] - bits->offset;
      if (bits->width != 0)
      {
         bits->rowWIB = range->rowWIB;
         bits->bitmapTable = range->bitmapTable;
      }
   }
}
static void destroyUff(struct UserFont *uf)
{
   struct UffUserFont *uff = uf->uff;
   if (uff->ranges) // guich@552_14: check against out of memory on this member
   {
	   int16 nParts = uff->nParts;
	   while (--nParts >= 0)
	   {
	      xfree(uff->ranges[nParts].bitIndexTable);
	      if (uff->ranges[nParts].recH)
	      {
	         MemHandleUnlock(uff->ranges[nParts].recH);
	      }
	   }
	}
   xfree(uf->fontP);
   xfree(uff->recNo);
   xfree(uff->uffFirstChar);
   xfree(uff->uffLastChar);
   xfree(uff->ranges);
   xfree(uf->uff);
   //can't release uf->recH!
   if (uf->dmRef)
   {
      DmCloseDatabase(uf->dmRef);
   }
}

static int loadUff(DmOpenRef dmRef, struct UserFont*uf)
{
   VoidHand recH;
   uint16 nParts;
   byte *recPtr;
   uint16 *recNo,*fchar,*lchar;
   struct UffUserFont *uff = NULL;
   FontPtr f;

   uf->dmRef = dmRef;

   // load header record -  the name's already stored
   recH = DmQueryRecord(dmRef, 0);
   recPtr = (byte*)MemHandleLock(recH);
   recPtr += 2; // magic is 12003
   recPtr += 2+READ_BE_UINT16(recPtr); // fcName is the short name, e.g. "gnuuni", "myarial"
   recPtr += 2; // fcSize is the point size, e.g. 10, 12
   recPtr += 2; // fcSlant is 0 for normal font; 1, 2, 3 for bold, italic, bold-italic
   recPtr += 2+READ_BE_UINT16(recPtr); // comment
   uf->fontP = f = (FontPtr)vmGlobals->xmalloc(SIZEOF_FONTTYPE);
   uf->uff = NULL;
   xmemzero(f,SIZEOF_FONTTYPE);
   if (f != NULL)
   {
      f->fontType     = READ_BE_UINT16(recPtr);          recPtr += 2;
      f->maxWidth     = READ_BE_UINT16(recPtr);          recPtr += 2;
      f->kernMax      = READ_BE_UINT16(recPtr);          recPtr += 2;
      f->nDescent     = READ_BE_UINT16(recPtr);          recPtr += 2;
      f->fRectWidth   = READ_BE_UINT16(recPtr);          recPtr += 2;
      f->fRectHeight  = READ_BE_UINT16(recPtr);          recPtr += 2;
      f->ascent       = READ_BE_UINT16(recPtr);          recPtr += 2;
      f->descent      = READ_BE_UINT16(recPtr);          recPtr += 2;
      f->leading      = READ_BE_UINT16(recPtr);          recPtr += 2;
      // versus UserFonts, we miss owTLoc and rowWords
      uf->uff           = uff = (struct UffUserFont*)vmGlobals->xmalloc(sizeof(struct UffUserFont));
   }
   if (uf->uff == NULL)
   {
      MemHandleUnlock(recH);
      DmCloseDatabase(dmRef);
      return 0;
   }
   xmemzero(uf->uff,sizeof(struct UffUserFont));
   uff->nParts       = nParts = READ_BE_UINT16(recPtr); recPtr += 2;

   uff->recNo        = recNo = (uint16 *)vmGlobals->xmalloc(nParts<<1);
   uff->uffFirstChar = fchar = (uint16 *)vmGlobals->xmalloc(nParts<<1);
   uff->uffLastChar  = lchar = (uint16 *)vmGlobals->xmalloc(nParts<<1);
   uff->ranges       = (struct UserFontRange *)vmGlobals->xmalloc(nParts*sizeof(struct UserFontRange));
   if (recNo == NULL || fchar == NULL || lchar == NULL || uff->ranges == NULL)
   {
      MemHandleUnlock(recH); // guich@552_14: also unlock the font
      destroyUff(uf);
      vmGlobals->debug(vmGlobals->int2str("The UFF font has too many range parts and cannot be loaded. The max allowed is 4093 but it has",nParts)); // guich@552_14
      return 0;
   }
   xmemzero(uff->ranges,nParts*sizeof(struct UserFontRange));

   while (nParts-- > 0)
   {
      *recNo++ = READ_BE_UINT16(recPtr);  recPtr += 2;
      *fchar++ = (uint16)READ_BE_UINT32(recPtr);  recPtr += 4; // not sure why the original version is not 16 bits
      *lchar++ = (uint16)READ_BE_UINT32(recPtr);  recPtr += 4;
   }

   uf->recH = (VoidHand)1; // can't be 0, otherwise the font won't be displayed - guich@552_14: moved to here
   MemHandleUnlock(recH);
   return 1;
}
//
///////////////////////////////////////////////////////////
int nrFonts;
void releaseAllFonts()
{
   int i;
   struct UserFont *uf;
   for (i=nrFonts-1; i >= 0; i--)
      if ((uf=&vmGlobals->allFonts[i])->recH)
      {
         if (IS_UNICODE_FONT(uf))
            destroyUff(uf);
         else
         {
            MemHandleUnlock(uf->recH);
            //DmReleaseRecord(uf->dmRef, uf->recIndex, false); // guich@340_47 - guich@400_26: now we do a queryrecord; no need to release the record
            if (uf->recIndex == 0) DmCloseDatabase(uf->dmRef);
         }
      }
   if (vmGlobals->allFonts) xfree(vmGlobals->allFonts);
}

#define FM_STRINGWIDTH 1
#define FM_CHARARRAYWIDTH 2
#define FM_CHARWIDTH 3
int charWidth(int16 ch, struct UserFont *uf);
int charWidth(int16 ch, struct UserFont *uf)
{
   int16 index,width;
   if (IS_UNICODE_FONT(uf)) // guich@402_50: ufolib?
      return uffCharWidth(ch, uf);

   if (ch > 0xFF) ch = 0; // eisvogel@421_44: ignore unicode chars.
   index = ch - uf->fontP->firstChar;
   if (ch == 13) // guich@341_9
      return 0;
   if (index < 0 || ch > uf->fontP->lastChar)
   {
      // guich@320_50: fix support for invalid chars - index = numberOfChars; // MCS index
      return UF_MCS_WIDTH; //guich@401_12: directly return the value.  index = uf->numberOfChars;//guich@340_46 - (int)('?' - uf->fontP->firstChar);
   }
#if defined(PALMOS)
   if (uf->recH == NULL) // internal font? use Palm OS routines
   {
      uint16 density = vmGlobals->density;

      if (FntGetFont() != uf->fid && uf->fid <= SWLARGE_BOLD_FONT)
      {
         static FontID palmFontIDs[4] = {stdFont,boldFont,largeFont,largeBoldFont};
         FntSetFont(palmFontIDs[uf->fid]);
      }
      if (density != kCoordinatesStandard) WinSetCoordinateSystem(density);
      width = FntWCharWidth(ch);
      if (density != kCoordinatesStandard) WinSetCoordinateSystem(kCoordinatesStandard);
   }
   else
#endif
   {
      width = uf->bitIndexTable[index+1] - uf->bitIndexTable[index];
   }
   if (width == 0) // guich@340_46
      width = UF_MCS_WIDTH;

   return width;
}

static void FontMetricsGetWidth(int type, Var stack[])
{
   WObject font, fontMetrics; // guich@450_36: surface is ignored from now on
   int32 width=0;
   struct UserFont *uf;

   fontMetrics = stack[0].obj;
   font = WOBJ_FontMetricsFont(fontMetrics);
   if (font == 0)
   {
      RETURN32(0);
   }
   // surface is unused and user fonts are supported!
   uf = WOBJ_FontUserFont(font);

   switch (type)
   {
      case FM_CHARWIDTH:
      {
         width = (int32)charWidth((int16)stack[1].intValue,uf);
         break;
      }
      case FM_STRINGWIDTH:
      case FM_CHARARRAYWIDTH:
      {
         WObject string, charArray;
         int32 start, count;
         uint16 *chars;

         width = 0;
         if (type == FM_STRINGWIDTH)
         {
            string = stack[1].obj;
            if (string == 0)
               break;
            charArray = WOBJ_StringCharArrayObj(string);
            if (charArray == 0)
               break;
            start = 0;
            count = WOBJ_arrayLen(charArray);
         }
         else // FM_CHARARRAYWIDTH
         {
            charArray = stack[1].obj;
            start = stack[2].intValue;
            count = stack[3].intValue;
            if ((*vmGlobals->arrayRangeCheck)(charArray, start, count) == 0)
               break; // array null or range invalid
         }
         chars = (uint16 *)WOBJ_arrayStart(charArray);
         chars = &chars[start];
         while (count-- > 0)
            width += (int32)charWidth(*chars++, uf); // guich@200b2: must handle unicode chars in the future
         _LOG_DEBUG((LOG_CAT_FONT, "FontMetricsGetWidth(%s)=%d", &chars[start], width)); // TODO@ string is unicode
         break;
      }
   }
   RETURN32(width);
}

static int getUserFontIndex(char *name, char *sufix)
{
   int i,len;
   char buf[41];
   xmemzero(buf,sizeof(buf));
   len = xstrlen(name);
   xstrncpy(buf,name,len);
   if (sufix) xstrncpy(buf+len,sufix,xstrlen(sufix));
   for (i =0; i < nrFonts; i++)
      if (vmGlobals->allFonts[i].name && !StrCaselessCompare(buf,vmGlobals->allFonts[i].name))
         return i;
   return -1;
}

static FontPtr getFontPtr(FontID fontID)
{
   FontID prevFontID;
   FontPtr fontPtr;
   // sets the main font
   prevFontID = FntSetFont(fontID);
   // gets the current font ptr
   fontPtr = FntGetFontPtr();
   _LOG_DEBUG((LOG_CAT_FONT, "getFontPtr(%d)=0x%x", fontID, fontPtr));

   // since we dont use the standard routines, restore the original font
   FntSetFont(prevFontID);
   return fontPtr;
}

static void storeUserFont(struct UserFont *uf, FontPtr fontP, char *name, VoidHand recH, DmOpenRef dmRef,
							uint16 recIndex, uint16 idx)
{
   uint32 bitmapTableSize;
   uint16 *bit;
   int i;
#if SDL_BYTEORDER == SDL_LIL_ENDIAN
   FontTypeV2 *fontV2=0;
#endif

   xstrncpy(uf->name,name,min(xstrlen(name),40));
   _LOG_INFO((LOG_CAT_FONT, "storeUserFont(%s)", name));

   uf->recH = recH;
   uf->fid = idx;
   uf->dmRef = dmRef;

   #if SDL_BYTEORDER == SDL_LIL_ENDIAN
   if (recH != NULL) // font file
   {
      _LOG_DEBUG((LOG_CAT_FONT, "DO_SWAP16"));
      FontPtr f = (FontPtr)vmGlobals->xmalloc(SIZEOF_FONTTYPE);
      if (f != NULL)
      {
         uint16 *recPtr = (uint16 *)fontP;
         xmemzero(f,SIZEOF_FONTTYPE);
         f->fontType     = READ_BE_UINT16(recPtr);          recPtr++;
         f->firstChar    = READ_BE_UINT16(recPtr);          recPtr++;
         f->lastChar     = READ_BE_UINT16(recPtr);          recPtr++;
         f->maxWidth     = READ_BE_UINT16(recPtr);          recPtr++;
         f->kernMax      = READ_BE_UINT16(recPtr);          recPtr++;
         f->nDescent     = READ_BE_UINT16(recPtr);          recPtr++;
         f->fRectWidth   = READ_BE_UINT16(recPtr);          recPtr++;
         f->fRectHeight  = READ_BE_UINT16(recPtr);          recPtr++;
         f->owTLoc       = READ_BE_UINT16(recPtr);          recPtr++;
         f->ascent       = READ_BE_UINT16(recPtr);          recPtr++;
         f->descent      = READ_BE_UINT16(recPtr);          recPtr++;
         f->leading      = READ_BE_UINT16(recPtr);          recPtr++;
         f->rowWords     = READ_BE_UINT16(recPtr);          recPtr++;
         uf->fontP = f;
      }
   }
   else // internal font
   {
//#if USE_INTERNAL_FONT == 1
      if (vmGlobals->screenWidth==160)
      {
         _LOG_DEBUG((LOG_CAT_FONT, "DONT_SWAP16"));
         uf->fontP = fontP;
      }
      else
      {
         _LOG_DEBUG((LOG_CAT_FONT, "X2"));
         FontPtr f = (FontPtr)vmGlobals->xmalloc(SIZEOF_FONTTYPE);
         if (f != NULL)
         {
            uint16 *recPtr = (uint16*)fontP;
            xmemzero(f,SIZEOF_FONTTYPE);
            f->fontType     = *recPtr;          recPtr++;
            f->firstChar    = *recPtr;          recPtr++;
            f->lastChar     = *recPtr;          recPtr++;
            f->maxWidth     = *recPtr << 1;     recPtr++;
            f->kernMax      = *recPtr;          recPtr++;
            f->nDescent     = *recPtr;          recPtr++;
            f->fRectWidth   = *recPtr << 1;     recPtr++;
            f->fRectHeight  = *recPtr << 1;     recPtr++;
            f->owTLoc       = *recPtr;          recPtr++;
            f->ascent       = *recPtr << 1;     recPtr++;
            f->descent      = *recPtr << 1;     recPtr++;
            f->leading      = *recPtr << 1;     recPtr++;
            f->rowWords     = *recPtr << 1;     recPtr++;
            uf->fontP = f;
         }
      }
//#endif
   }
   #else
   _LOG_DEBUG((LOG_CAT_FONT, "DONT_SWAP16"));
   uf->fontP = fontP;
   #endif

   _LOG_DEBUG((LOG_CAT_FONT, "sizeof(FONTTYPE)=%d", (int)SIZEOF_FONTTYPE));
   _LOG_DEBUG((LOG_CAT_FONT, "fontType=0x%x", (int)uf->fontP->fontType));
   _LOG_DEBUG((LOG_CAT_FONT, "firstChar=0x%x", (int)uf->fontP->firstChar));
   _LOG_DEBUG((LOG_CAT_FONT, "lastChar=0x%x", (int)uf->fontP->lastChar));
   _LOG_DEBUG((LOG_CAT_FONT, "maxWidth=0x%x", (int)uf->fontP->maxWidth));
   _LOG_DEBUG((LOG_CAT_FONT, "fRectHeight=0x%x", (int)uf->fontP->fRectHeight));
   _LOG_DEBUG((LOG_CAT_FONT, "rowWords=0x%x", (int)uf->fontP->rowWords));

   uf->recIndex = recIndex;
   uf->rowWidthInBytes = ((uint32)uf->fontP->rowWords) * 2;
   _LOG_DEBUG((LOG_CAT_FONT, "uf->rowWidthInBytes=0x%x", (int)uf->rowWidthInBytes));

   uf->numberOfChars = uf->fontP->lastChar - uf->fontP->firstChar + 1;
   _LOG_DEBUG((LOG_CAT_FONT, "uf->numberOfChars=0x%x", (int)uf->numberOfChars));

#if SDL_BYTEORDER == SDL_LIL_ENDIAN
   if (recH == NULL) // internal font
   {
      fontV2 = (FontTypeV2*)fontP;
      _LOG_DEBUG((LOG_CAT_FONT, "densityCount=%d", (int)fontV2->densityCount));
      for (i=0; i < fontV2->densityCount; i++)
      {
         _LOG_DEBUG((LOG_CAT_FONT, "density=0x%x", (int)fontV2->densities[i].density));
         _LOG_DEBUG((LOG_CAT_FONT, "glyphBitsOffset=0x%x", (int)fontV2->densities[i].glyphBitsOffset));
      }
      uf->bitmapTable = (byte*)fontP + fontV2->densities[/*USE_INTERNAL_FONT-1*/(vmGlobals->screenWidth==160)?0:1].glyphBitsOffset;
   }
   else
   {
      uf->bitmapTable = (byte*)fontP + SIZEOF_FONTTYPE;
   }
#else
   uf->bitmapTable = (byte*)fontP + SIZEOF_FONTTYPE;
#endif
   _LOG_DEBUG((LOG_CAT_FONT, "uf->bitmapTable=0x%x", (int)uf->bitmapTable));

   bitmapTableSize = ((uint32)uf->rowWidthInBytes) * uf->fontP->fRectHeight;
   _LOG_DEBUG((LOG_CAT_FONT, "bitmapTableSize=0x%x", (int)bitmapTableSize));

   //dumpHex(fontP, 8);

#if SDL_BYTEORDER == SDL_LIL_ENDIAN
   if (recH != NULL) // font file
   {
      uf->bitIndexTable = (int16*)vmGlobals->xmalloc(uf->numberOfChars * sizeof(int16));
      bit = (int16 *)(uf->bitmapTable + bitmapTableSize);

      // invert the bit indexes
      for (i = 0; i < uf->numberOfChars; i++)
      {
         uf->bitIndexTable[i] = READ_BE_UINT16(&bit[i]);
         if (i < 5)
         {
            _LOG_DEBUG((LOG_CAT_FONT, "inv bitidx=0x%x", (int)uf->bitIndexTable[i]));
      	 }
      }
   }
   else // internal font
   {
//#if USE_INTERNAL_FONT == 1
      if (vmGlobals->screenWidth==160)
      {
         //uf->bitIndexTable = (int16*)(uf->bitmapTable + bitmapTableSize + 58);
         uf->bitIndexTable = (int16*) &fontV2->densities[2];
         for (i = 0; i < uf->numberOfChars; i++)
         {
            if (i < 10 || i > (uf->numberOfChars-10))
            {
               _LOG_DEBUG((LOG_CAT_FONT, "ninv bitidx=0x%x", (int)uf->bitIndexTable[i]));
            }
         }
      }
      else
      {
         uf->bitIndexTable = (int16*)vmGlobals->xmalloc(uf->numberOfChars * sizeof(int16));
         bit = (int16 *)&fontV2->densities[2];

         // fix the bit indexes
         for (i = 0; i < uf->numberOfChars; i++)
         {
            uf->bitIndexTable[i] = bit[i] << 1;
            if (i < 10 || i > (uf->numberOfChars-10))
            {
               _LOG_DEBUG((LOG_CAT_FONT, "fix bitidx=0x%x", (int)uf->bitIndexTable[i]));
            }
         }
      }
//#endif

   }
#else
   uf->bitIndexTable = (int16*)(uf->bitmapTable + bitmapTableSize);
#endif

   _LOG_DEBUG((LOG_CAT_FONT, "uf->bitIndexTable=0x%x", (int)uf->bitIndexTable));
}

byte loadAllFonts()
{
   DmSearchStateType state;
   DmOpenRef dmRef;
   Boolean first;
   UInt cardNo = 0;
   LocalID dbID;
   char dbName[41]; // guich@340_47
   Err err;
   ULong crid = REGFONT_CREATOR_ID;
   ULong cridu = UFOFONT_CREATOR_ID;
   ULong typeu = UFOFONT_TYPE_ID;
   int installed=0;
   int unicodeInstalled = 0;
   int i=0;
   int internal=4;
   int index = 0, specialIndex; // guich@400_41: fixed problem with storing internal fonts

   // first find how many installed fonts
   first = true;
   while (1)
   {
      uint32 numRecords;
      _LOG_DEBUG((LOG_CAT_FONT, "DmGetNextDatabaseByTypeCreator=0x%x", (int)crid));
      err = DmGetNextDatabaseByTypeCreator(first, &state, crid, crid, false, PALM_CARD(&cardNo) &dbID);
      if (err == dmErrCantFind)
      {
         break;
      }
      // guich@401_23: always delete the 5SW file from now on
      if (DmDatabaseInfo(PALM_CARD(cardNo) dbID, dbName, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL) == 0
      		            && dbName[0] == '5' && xstrncmp(dbName,"5SW",3) == 0) // guich@400_24
      {
         _LOG_DEBUG((LOG_CAT_FONT, "Delete %s", dbName));
         DmDeleteDatabase(PALM_CARD(cardNo) dbID);
      }
      else
      {
         if (DmDatabaseSize(PALM_CARD(cardNo) dbID, &numRecords, NULL, NULL) == errNone) // guich@340_47
         {
            installed += numRecords;
            _LOG_DEBUG((LOG_CAT_FONT, "%d more fonts", numRecords));
         }
         else
         {
            installed++;
            _LOG_DEBUG((LOG_CAT_FONT, "one more fonts"));
         }
      }
      first = false;
   }
   // guich@402_50: now finds how many unicode fonts are installed
   first = true;
   while (1)
   {
      _LOG_DEBUG((LOG_CAT_FONT, "DmGetNextDatabaseByTypeCreator=0x%x", (int)cridu));
      err = DmGetNextDatabaseByTypeCreator(first, &state, typeu, cridu, false, PALM_CARD(&cardNo) &dbID);
      if (err == dmErrCantFind)
      {
      	break;
      }
      unicodeInstalled++; // unicode fonts has many records inside it, but count as only one font
      first = false;
   }
   // alloc mem
   nrFonts = unicodeInstalled + installed + internal;
   _LOG_DEBUG((LOG_CAT_FONT, "FONTS=(%d, %d, %d)=%d", (int)unicodeInstalled, (int)installed, (int)internal, (int)nrFonts));

   vmGlobals->allFonts = (struct UserFont *) (*vmGlobals->xmalloc)(sizeof(struct UserFont)*nrFonts); // important! allFonts can't be changed at runtime, bc fonts are directly mapped to it.
   xmemzero(vmGlobals->allFonts,sizeof(struct UserFont)*nrFonts);

#define hrStdFont       0
#define hrBoldFont      0
#define hrLargeFont     0
#define hrLargeBoldFont 0

//#if USE_INTERNAL_FONT  - guich: always use it
   // get the internal fonts
   storeUserFont(vmGlobals->allFonts + SWSTANDARD_FONT,
                getFontPtr(vmGlobals->isSony?hrStdFont:stdFont),"SWSmall", 0, 0, 0, SWSTANDARD_FONT);
   storeUserFont(vmGlobals->allFonts + SWSTANDARD_BOLD_FONT,
                getFontPtr(vmGlobals->isSony?hrBoldFont:boldFont),"SWSmallBold", 0, 0, 0, SWSTANDARD_BOLD_FONT);
   if (vmGlobals->romVersion >= 0x03000000L) // has large fonts?
   {
      _LOG_DEBUG((LOG_CAT_FONT, "large fonts"));
      storeUserFont(vmGlobals->allFonts + SWLARGE_FONT,
                    getFontPtr(vmGlobals->isSony?hrLargeFont:largeFont),"SWLarge", 0, 0, 0, SWLARGE_FONT);
      storeUserFont(vmGlobals->allFonts + SWLARGE_BOLD_FONT,
                    getFontPtr(vmGlobals->isSony?hrLargeBoldFont:largeBoldFont),"SWLargeBold", 0, 0, 0, SWLARGE_BOLD_FONT);
   }
   else
   {
      _LOG_DEBUG((LOG_CAT_FONT, "no large fonts, vmGlobals->romVersion=0x%x", (int)vmGlobals->romVersion));
   }
//#endif

   // now get the installed fonts
   first = true;
   for (i = 0; i < installed;)
   {
      uint16 attr;
      int idx;
      _LOG_DEBUG((LOG_CAT_FONT, "DmGetNextDatabaseByTypeCreator=0x%x", (int)crid));
      err = DmGetNextDatabaseByTypeCreator(first, &state, crid, crid, false, PALM_CARD(&cardNo) &dbID);
      if (DmDatabaseInfo(PALM_CARD(cardNo) dbID, dbName, &attr, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL) == 0)
      {
         VoidHand recH;
         CharPtr recPtr;
         char *dbAlias = dbName;
         uint32 recCount; // guich@340_47
         uint32 j;

         _LOG_DEBUG((LOG_CAT_FONT, "font=%s", dbName));

         dmRef = DmOpenDatabase(PALM_CARD(cardNo) dbID, dmModeReadOnly);
         _LOG_DEBUG((LOG_CAT_FONT, " dmRef=%x", (int)dmRef));

         recCount = DmNumRecords(dmRef);
         _LOG_DEBUG((LOG_CAT_FONT, " recCount=%d", (int)recCount));
         for (j = 0; j < recCount; j++) // guich@340_47
         {
            recH = DmQueryRecord(dmRef, j); // guich@400_26
            recPtr = MemHandleLock(recH);
            _LOG_DEBUG((LOG_CAT_FONT, " recPtr=0x%x", (int)recPtr));
            if (recCount > 1) // packaged font file? get the name
            {
               int len;
               _LOG_DEBUG((LOG_CAT_FONT, " packaged font file? name=%s", (char *)recPtr));

               len = xstrlen((char *)recPtr)+1;
               if (len < 4) // bad font file?
               {
                  _LOG_ERROR((LOG_CAT_FONT, "bad font file"));
                  MemHandleUnlock(recH);
                  //DmReleaseRecord(dmRef, j, false); guich@402_51
                  DmCloseDatabase(dmRef);
                  break;
               }
               len = (int)((len+3) / 4)*4; // 4 byte align
               dbAlias = (char *)recPtr;
               recPtr += len; // point to the font structure
               _LOG_DEBUG((LOG_CAT_FONT, " recPtr=0x%x", (int)recPtr));
            }
            if (*(dbAlias+1) == 'S' && *(dbAlias+2) == 'W')
            {
               dbAlias++;
            }
            specialIndex = -1;
            if (!StrCompare("SWSmall", dbAlias))
            {
               specialIndex = SWSTANDARD_FONT;
           	}
           	else if (!StrCompare("SWSmallBold", dbAlias))
           	{
               specialIndex = SWSTANDARD_BOLD_FONT;
            }

            if (!StrCompare("SWLarge", dbAlias))
            {
               specialIndex = SWLARGE_FONT;
            }
            else if (!StrCompare("SWLargeBold", dbAlias))
            {
               specialIndex = SWLARGE_BOLD_FONT;
            }
           	idx = (specialIndex != -1) ? specialIndex : (internal + index++);
           	storeUserFont(&vmGlobals->allFonts[idx], (FontPtr)recPtr, dbAlias, recH, dmRef, j, idx);
       	 }
         i += recCount;
	  }
      first = false;
   }

   // get the unicode ones
   first = true;
   index += internal;
   while (1)
   {
      _LOG_DEBUG((LOG_CAT_FONT, "DmGetNextDatabaseByTypeCreator=0x%x", (int)cridu));
      err = DmGetNextDatabaseByTypeCreator(first, &state, typeu, cridu, false, PALM_CARD(&cardNo) &dbID);
      if (err == dmErrCantFind) break;
      DmDatabaseInfo(PALM_CARD(cardNo) dbID, (char*)&vmGlobals->allFonts[index].name, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
      dmRef = DmOpenDatabase(PALM_CARD(cardNo) dbID, dmModeReadOnly);
      loadUff(dmRef,vmGlobals->allFonts+index++);
   	  first = false;
   }
   return 1;
}
//////////////////////////////////////////////////////////////////////////
// waba/fx/Font static void fontCreate(Font obj)
void FontCreate(Var stack[]) // (Lwaba/fx/Font;)V
{
  	WObject obj = stack[0].obj; // class object
   int bold = WOBJ_FontStyle(obj) == Font_BOLD;
   int large = WOBJ_FontSize(obj) >= 14;
  	UtfString fontName;
 	int index;
   char *c;

   //static char *sufixes[4] = {"Small","SmallBold","Large","LargeBold"};
   char *sufix = 0;
   fontName = (*vmGlobals->stringToUtf)(WOBJ_FontName(obj), STU_USE_STATIC | STU_NULL_TERMINATE);
   c = xstrstr(fontName.str,"Small");
   if (c != NULL && c != fontName.str)
   {
      *c = 0;
   }
   c = xstrstr(fontName.str,"Large");
   if (c != NULL && c != fontName.str)
   {
      *c = 0;
   }

   if (fontName.len > 3 && (fontName.str[1] == fontName.str[2]) && (fontName.str[0] == 'U' || fontName.str[1] == 'u') && (fontName.str[1] == 'F' || fontName.str[1] == 'f')) // is it uff?
   {
      sufix = (vmGlobals->density == kCoordinatesStandard)?"_L":"_H";
   }
   else
   {
      switch (large*2+bold)
      {
         case 0: sufix = "Small";     break;
         case 1: sufix = "SmallBold"; break;
         case 2: sufix = "Large";     break;
         case 3: sufix = "LargeBold"; break;
      }
   }

   // first, try to load the font itself using the sw font pattern
   index = getUserFontIndex(fontName.str,sufix);
   if (index == -1)
   {
      // ok, the pattern didnt work, use the font name given
      if (!(xstrlen(fontName.str) > 2 && fontName.str[0] == 'S' && fontName.str[1] == 'W')) // is it the internal ones? dont try the name itself
      {
         index = getUserFontIndex(fontName.str,NULL);
      }
      if (index == -1)
      {
         // ok, the font is not found, use the internal ones
         index = getUserFontIndex(/*(MaxScreenX>=320)?(vmGlobals->isSony?"HSW":"5SW"):(MaxScreenX>=240)?"MSW":"LSW"*/"SW",sufix);
         if (index == -1)
         {
            index = bold?1:0; // internal ones not found (maybe large in palmOS 2.0), use the default small font.
         }
         // change the font name
         WOBJ_FontName(obj) = (*vmGlobals->createString)(vmGlobals->allFonts[index].name);
      }
   }
   WOBJ_FontUserFont(obj) = vmGlobals->allFonts+index;
}
//////////////////////////////////////////////////////////////////////////
// waba/fx/FontMetrics static void fontMetricsCreate(FontMetrics obj)
void FontMetricsCreate(Var stack[]) // (Lwaba/fx/FontMetrics;)V
{
   WObject font, fontMetrics;
   struct UserFont *uf;

   fontMetrics = stack[0].obj;
   font = WOBJ_FontMetricsFont(fontMetrics);
   if (font == 0)
   {
      WOBJ_FontMetricsAscent(fontMetrics) = 0;
      WOBJ_FontMetricsDescent(fontMetrics) = 0;
      WOBJ_FontMetricsLeading(fontMetrics) = 0;
      RETURN32(0);
   }
   // surface is unused and user defined fonts are supported!
   uf = WOBJ_FontUserFont(font);

   WOBJ_FontMetricsAscent(fontMetrics) = uf->fontP->ascent;
   WOBJ_FontMetricsDescent(fontMetrics) = uf->fontP->descent;
   WOBJ_FontMetricsLeading(fontMetrics) = uf->fontP->leading;
}
//////////////////////////////////////////////////////////////////////////
// waba/fx/FontMetrics static int fontMetricsGetCharWidth(FontMetrics obj, char c)
void FontMetricsGetCharWidth(Var stack[]) // (Lwaba/fx/FontMetrics;C)I
{
   FontMetricsGetWidth(FM_CHARWIDTH, stack);
}
//////////////////////////////////////////////////////////////////////////
// waba/fx/FontMetrics static int fontMetricsGetTextWidth(FontMetrics obj, String s)
void FontMetricsGetTextWidth_String(Var stack[]) // (Lwaba/fx/FontMetrics;Ljava/lang/String;)I
{
   FontMetricsGetWidth(FM_STRINGWIDTH, stack);
}
//////////////////////////////////////////////////////////////////////////
// waba/fx/FontMetrics static int fontMetricsGetTextWidth(FontMetrics obj, char chars[], int start, int count)
void FontMetricsGetTextWidth(Var stack[]) // (Lwaba/fx/FontMetrics;[CII)I
{
   FontMetricsGetWidth(FM_CHARARRAYWIDTH, stack);
}

#else // defined PALMOS  ----------------------------------------------------------------------

///////////////////////////////////////////////////////////
//          New Unicode Support

// Load and cache used ranges. pending: unload unused ranges
// Return index in range arrays, or -1 if MCS
static int32 uffEnsureRange(char16 ch, struct UserFont *uf)
{
   struct UffUserFont *uff = uf->uff;
   // check if the range is already in the cache - if so, it is sure it is already loaded.
   if (ch <= uff->recentLastChar && ch >= uff->recentFirstChar)
      return uff->recentRangeIndex;
   else
   {
      int32 rindex = -1;

      // Find range. Perform a binary search here. This whould help if a lot of ranges.
      int32 low = 0;
      int32 hi = uff->nParts - 1;
      int32 mid = 0;

      while (low <= hi)
      {
         mid = (low + hi) >> 1;
         if (ch >= uff->uffFirstChar[mid] && ch <= uff->uffLastChar[mid])
         {
            rindex = mid;
            break;
         }
         else
         if (ch < uff->uffFirstChar[mid])
            hi = mid - 1;
         else
            low = mid + 1;
      }

      if (rindex != -1)
      {
         if (uff->ranges[rindex].rowWIB == 0) // not yet loaded
         {
            int32 size;
            struct UffUserFont *uff = uf->uff;
            struct UserFontRange *range = &uff->ranges[rindex];
            byte *recPtr;
            int32 *bit;
            range->recH = DmQueryRecord(uf->dmRef, uff->recNo[rindex],0); // load record
            recPtr = (byte*)MemHandleLock(range->recH);
            range->rowWIB = READ_BE_UINT32(recPtr); recPtr += 4;

            size = (int32)range->rowWIB * (int32)uf->fontP->fRectHeight;
            range->bitmapTable = recPtr; recPtr += size;

            // unfortunately, we can't point to the bitIndexTable due to memory alinment
            size = uff->uffLastChar[rindex] - uff->uffFirstChar[rindex] + 1 + 1;
            range->bitIndexTable = bit = (int32*)vmGlobals->xmalloc(size<<2);
            if (bit == NULL)
            {
               xfree/*MemHandleUnlock*/(range->recH);
               return -1;
            }
            for (; size--; recPtr+=4) {
               *bit++ = READ_BE_INT32(recPtr);
            }
         }
         uff->recentFirstChar  = uff->uffFirstChar[rindex];
         uff->recentLastChar   = uff->uffLastChar[rindex];
         uff->recentRangeIndex = rindex;
      }
      return rindex;
   }
}

// Get the width of a character
static int uffCharWidth(char16 ch, struct UserFont *uf)
{
   struct UffUserFont *uff = uf->uff;
   // ensure that range is loaded
   int32 rindex = uffEnsureRange(ch,uf);
   if (rindex == -1)
      return uf->fontP->maxWidth>>1; // guich@402_55
   else
   {
      struct UserFontRange *range = &uff->ranges[rindex];
      int32 index = (int32)ch - uff->recentFirstChar;
      return range->bitIndexTable[index + 1] - range->bitIndexTable[index];
   }
}

// Update a CharBits object to reflect the specificities of a given character.
void uffSetCharBits(char16 ch, struct UserFont *uf, struct CharBits *bits)
{
   struct UffUserFont *uff = uf->uff;
   int32 rindex = uffEnsureRange(ch,uf);
   if (rindex == -1)
      bits->width = 0;
   else
   {
      int32 cindex = (int32) ch - uff->recentFirstChar;
      struct UserFontRange *range = &uff->ranges[rindex];
      bits->offset = range->bitIndexTable[cindex];
      bits->width = range->bitIndexTable[cindex + 1] - bits->offset;
      if (bits->width != 0)
      {
         bits->rowWIB = range->rowWIB;
         bits->bitmapTable = range->bitmapTable;
      }
   }
}
static void destroyUff(struct UserFont *uf)
{
   struct UffUserFont *uff = uf->uff;
   int16 nParts = uff->nParts;
   while (--nParts >= 0)
   {
      xfree(uff->ranges[nParts].bitIndexTable);
      if (uff->ranges[nParts].recH)
         xfree/*MemHandleUnlock*/(uff->ranges[nParts].recH);
      }
   xfree(uf->fontP);
   xfree(uff->recNo);
   xfree(uff->uffFirstChar);
   xfree(uff->uffLastChar);
   xfree(uff->ranges);
   xfree(uf->uff);
   //can't release uf->recH!
   if (uf->dmRef)
      DmCloseDatabase(uf->dmRef);
}

static int loadUff(DmOpenRef dmRef, struct UserFont*uf)
{
   VoidHand recH;
   uint16 nParts;
   byte *recPtr;
   uint16 *recNo,*fchar,*lchar;
   struct UffUserFont * uff = NULL;
   FontPtr f;

   uf->dmRef = dmRef;
   uf->recH = (VoidHand)1; // can't be 0, otherwise the font won't be displayed
   // load header record -  the name's already stored
   recH = DmQueryRecord(dmRef, 0,0);
   recPtr = (byte*)MemHandleLock(recH);
   recPtr += 2; // magic is 12003
   recPtr += 2+READ_BE_UINT16(recPtr); // fcName is the short name, e.g. "gnuuni", "myarial"
   recPtr += 2; // fcSize is the point size, e.g. 10, 12
   recPtr += 2; // fcSlant is 0 for normal font; 1, 2, 3 for bold, italic, bold-italic
   recPtr += 2+READ_BE_UINT16(recPtr); // comment
   f = (FontPtr)vmGlobals->xmalloc(sizeof(FontType));
// uf->uff = NULL;
   if (f)
   {
      xmemzero(f, sizeof(FontType));
      f->fontType     = READ_BE_UINT16(recPtr);          recPtr += 2;
      f->maxWidth     = READ_BE_UINT16(recPtr);          recPtr += 2;
      f->kernMax      = READ_BE_UINT16(recPtr);          recPtr += 2;
      f->nDescent     = READ_BE_UINT16(recPtr);          recPtr += 2;
      f->fRectWidth   = READ_BE_UINT16(recPtr);          recPtr += 2;
      f->fRectHeight  = READ_BE_UINT16(recPtr);          recPtr += 2;
      f->ascent       = READ_BE_UINT16(recPtr);          recPtr += 2;
      f->descent      = READ_BE_UINT16(recPtr);          recPtr += 2;
      f->leading      = READ_BE_UINT16(recPtr);          recPtr += 2;
      // versus UserFonts, we miss owTLoc and rowWords
      uff = (struct UffUserFont*)vmGlobals->xmalloc(sizeof(struct UffUserFont));
      if (!uff)
      {
         xfree(f);
         f = 0;
      }
   }
   if (!f)
   {
      xfree/*MemHandleUnlock*/(recH);
      DmCloseDatabase(dmRef);
      return 0;
   }
   uf->fontP = f;
   uf->uff = uff;
   xmemzero(uff,sizeof(struct UffUserFont));
   uff->nParts       = nParts = READ_BE_UINT16(recPtr); recPtr += 2;
   uff->recNo        = recNo = (uint16 *)vmGlobals->xmalloc(nParts<<1);
   uff->uffFirstChar = fchar = (uint16 *)vmGlobals->xmalloc(nParts<<1);
   uff->uffLastChar  = lchar = (uint16 *)vmGlobals->xmalloc(nParts<<1);
   uff->ranges       = (struct UserFontRange *)vmGlobals->xmalloc(nParts*sizeof(struct UserFontRange));
   if (recNo == NULL || fchar == NULL || lchar == NULL || uff->ranges == NULL)
   {
      destroyUff(uf);
      return 0;
   }
   xmemzero(uff->ranges,nParts*sizeof(struct UserFontRange));

   while (nParts-- > 0)
   {
      *recNo++ = READ_BE_UINT16(recPtr);  recPtr += 2;
      *fchar++ = (uint16)READ_BE_UINT32(recPtr);  recPtr += 4; // not sure why the original version is not 16 bits
      *lchar++ = (uint16)READ_BE_UINT32(recPtr);  recPtr += 4;
   }
   xfree/*MemHandleUnlock*/(recH);

   return 1;
}
//
///////////////////////////////////////////////////////////

int nrFonts;
#define FM_STRINGWIDTH 1
#define FM_CHARARRAYWIDTH 2
#define FM_CHARWIDTH 3

int charWidth(char16 ch, struct UserFont *uf);
int charWidth(char16 ch, struct UserFont *uf)
{
   int index = (int)ch - uf->fontP->firstChar;
   int width;
   if (IS_UNICODE_FONT(uf)) // guich@402_50: ufolib?
      return uffCharWidth(ch, uf);
   if (ch == 13) // guich@341_9
      return 0;
   if (index < 0 || ch > uf->fontP->lastChar)
   {
      // guich@320_50: fix support for invalid chars - index = numberOfChars; // MCS index
      return UF_MCS_WIDTH; //guich@401_12: directly return the value.  index = uf->numberOfChars;//guich@340_46 - (int)('?' - uf->fontP->firstChar);
   }
   width = uf->bitIndexTable[index+1] - uf->bitIndexTable[index];
   if (width == 0) // guich@340_46
      width = UF_MCS_WIDTH; //(uf->bitIndexTable[uf->numberOfChars+1] - uf->bitIndexTable[uf->numberOfChars]); // width of MCB
   return width;
}

static void FontMetricsGetWidth(int type, Var stack[])
{
   WObject font, fontMetrics; // guich@450_36: surface is ignored from now on
   int32 width=0;
   struct UserFont *uf;

   fontMetrics = stack[0].obj;
   font = WOBJ_FontMetricsFont(fontMetrics);
   if (font == 0)
      RETURN32(0)
   uf = WOBJ_FontUserFont(font);

   switch (type)
   {
      case FM_CHARWIDTH:
      {
         width = (int32)charWidth((char16)stack[1].intValue,uf);
         break;
      }
      case FM_STRINGWIDTH:
      case FM_CHARARRAYWIDTH:
      {
         WObject string, charArray;
         int32 start, count;
         uint16 *chars;

         width = 0;
         if (type == FM_STRINGWIDTH)
         {
            string = stack[1].obj;
            if (string == 0)
               break;
            charArray = WOBJ_StringCharArrayObj(string);
            if (charArray == 0)
               break;
            start = 0;
            count = WOBJ_arrayLen(charArray);
         }
         else // FM_CHARARRAYWIDTH
         {
            charArray = stack[1].obj;
            start = stack[2].intValue;
            count = stack[3].intValue;
            if ((*vmGlobals->arrayRangeCheck)(charArray, start, count) == 0)
               break; // array null or range invalid
         }
         chars = (uint16 *)WOBJ_arrayStart(charArray);
         chars = &chars[start];
         while (count-- > 0)
            width += (int32)charWidth((char16)*chars++, uf); // guich@200b2: must handle unicode chars in the future
         break;
      }
   }
   RETURN32(width)
}

static int getUserFontIndex(char *name, char *sufix)
{
   int i,len;
   char buf[41];
   xmemzero(buf,sizeof(buf));
   len = xstrlen(name);
   xstrncpy(buf,name,len);
   if (sufix) xstrncpy(buf+len,sufix,xstrlen(sufix));
   for (i =0; i < nrFonts; i++)
      if (vmGlobals->allFonts[i].name && !xstrcaselesscmp(buf,vmGlobals->allFonts[i].name))
         return i;
   return -1;
}

static void storeUserFont(struct UserFont *uf, FontPtr fontP, char *name, VoidHand recH, DmOpenRef dmRef, int recIndex, int len, int antialiased)
{
   uint16 i;

   #if SDL_BYTEORDER == SDL_LIL_ENDIAN
   /*
   | swap font member bytes
   | the previous #IF could be suppressed, b/c, for big endians,
   | swapIntXx is #define as SDL_SwapBE (Swap to Big Endians)
   */
   fontP->fontType = swapInt16(fontP->fontType);
   fontP->firstChar = swapInt16(fontP->firstChar);
   fontP->lastChar = swapInt16(fontP->lastChar);
   fontP->maxWidth = swapInt16(fontP->maxWidth);
   fontP->kernMax = swapInt16(fontP->kernMax);
   fontP->nDescent = swapInt16(fontP->nDescent);
   fontP->fRectWidth = swapInt16(fontP->fRectWidth);
   fontP->fRectHeight = swapInt16(fontP->fRectHeight);
   fontP->owTLoc = swapInt16(fontP->owTLoc);
   fontP->ascent = swapInt16(fontP->ascent);
   fontP->descent = swapInt16(fontP->descent);
   fontP->leading = swapInt16(fontP->leading);
   fontP->rowWords = swapInt16(fontP->rowWords);
   #endif

   xstrncpy(uf->name,name,min(xstrlen(name),40));
   uf->recH = (char *)recH - len; // guich@340_37
   uf->dmRef = dmRef;
   uf->fontP = fontP;
   uf->recIndex = (uint16)recIndex;
   if (antialiased) {
      // multiply by 4 (4 bits of transparency, vs single black or white bit)
      uf->rowWidthInBytes = uf->fontP->rowWords << 3;
      uf->uff = (void *)1;    /* PGR.  A bit hacky.  */
   }else {
      uf->rowWidthInBytes = uf->fontP->rowWords << 1;
   }
   uf->numberOfChars   = uf->fontP->lastChar - uf->fontP->firstChar + 1;
   uf->bitmapTable     = (byte*)uf->fontP + SIZEOF_FONTTYPE;
   uf->bitIndexTable   = (int16 *)(uf->bitmapTable + ((int32)uf->rowWidthInBytes * (int32)uf->fontP->fRectHeight));
   // invert the bit indexes
   #if SDL_BYTEORDER == SDL_LIL_ENDIAN
   for (i = 0; i <= uf->numberOfChars; i++)
      uf->bitIndexTable[i] = swapInt16(uf->bitIndexTable[i]);
   #endif
}

static byte countFonts(TCHAR *fileName, _unused_ void *nn)
{
   uint32 type, creator;
   uint16 count;
   DmGetDatabaseTypeCreatorCount(fileName, &type, &creator, &count);
   if (
      (type == REGFONT_TYPE_ID && creator == REGFONT_CREATOR_ID) ||
      (type == AALIASFONT_TYPE_ID && creator == AALIASFONT_CREATOR_ID)
   ) {
      nrFonts += count; // guich@340_47
   }
   else if (type == UFOFONT_TYPE_ID && creator == UFOFONT_CREATOR_ID) // guich@402_50
   {
      //PGR: Should we do a UAfn for UFO antialiased?
      ++nrFonts;
   }
   return 0;
}

static char *standardFontNames[] = {"SWSmall","SWSmallBold","SWLarge","SWLargeBold"};

static byte loadFonts(TCHAR *fileName, void *i)
{
   char dbName[41]; // guich@340_47
   char *dbAlias;
   int *ii = (int *)i;
   int index=0;
   int antialiased = 0;
   DmOpenRef dmRef;

   if (((dmRef = DmOpenDatabaseReadOnly(fileName, REGFONT_CREATOR_ID, REGFONT_TYPE_ID)) != NULL) ||
      (antialiased=1, (dmRef = DmOpenDatabaseReadOnly(fileName, AALIASFONT_CREATOR_ID, AALIASFONT_TYPE_ID)) != NULL))
   {
      uint16 recCount = DmNumRecords(dmRef); // guich@340_47
      uint16 j;
      int len;
      for (j = 0; j < recCount; j++) // guich@340_47
      {
         char *recH = DmQueryRecord(dmRef, j,0);
         if (recH == NULL)
         {
            DmSetRecordAttributes(dmRef, j,0); // this corrects a bug caused by the old version of Catalog at the desktop, that leaved the record busy
            recH = DmQueryRecord(dmRef, j,0);
            if (recH == NULL)
            {
               (*vmGlobals->debug)("Fonts not correctly installed!");
               return 1;
            }
         }
         len = 0;
         if (recCount > 1) // packaged font file? get the name
         {
            len = xstrlen((char *)recH)+1;
            if (len < 4) // bad font file?
            {
               xfree(recH);
               DmCloseDatabase(dmRef);
               break;
            }
            len = (int)((len+3) / 4)*4; // 4 byte align
            dbAlias = (char *)recH;
            recH += len; // point to the font structure
         }
         else
         {
            DmGetName(dmRef, dbName);
            dbAlias = dbName;
         }
         if ((dbAlias[1] == 'S') && (dbAlias[2] == 'W'))
         {
            ++dbAlias;
         }
         // don't change this comparison order
         if (xstrncmp(standardFontNames[SWSTANDARD_BOLD_FONT], dbAlias, 11) == 0) index = SWSTANDARD_BOLD_FONT; else
         if (xstrncmp(standardFontNames[SWSTANDARD_FONT],      dbAlias,  7) == 0) index = SWSTANDARD_FONT;      else
         if (xstrncmp(standardFontNames[SWLARGE_BOLD_FONT],    dbAlias, 11) == 0) index = SWLARGE_BOLD_FONT;    else
         if (xstrncmp(standardFontNames[SWLARGE_FONT],         dbAlias,  7) == 0) index = SWLARGE_FONT;
         else  // only increment if not one of the system fonts
         {
            index = *ii; // guich@400_65: compute index here
            *ii = *ii + 1;
         }
         storeUserFont(vmGlobals->allFonts + index, (FontPtr)recH, dbAlias, recH, dmRef,j,len,antialiased);
      }
   }
   else
   if ((dmRef = DmOpenDatabaseReadOnly(fileName, UFOFONT_CREATOR_ID,UFOFONT_TYPE_ID)) != NULL) // guich@402_50: is UFF?
   {
      index = *ii; // guich@400_65: compute index here
      *ii = *ii + 1;
      DmGetName(dmRef, (char*)&vmGlobals->allFonts[index].name);
      loadUff(dmRef, vmGlobals->allFonts + index);
   }
   return 0;
}

byte loadAllFonts()
{
   int i=4; // guich@400_65: start with 4

   // first, find how many installed fonts
   nrFonts = 0;
   DmListDatabases(&countFonts, NULL, 0); // every SW font is located under SUPERWABA_CORE_PATH
   if (nrFonts < 4)
   {
      showError(14001);  /*@@*/
      return 0;
   }

   // alloc mem
   vmGlobals->allFonts = (struct UserFont *) (*vmGlobals->xmalloc)(sizeof(struct UserFont)*nrFonts);
   xmemzero(vmGlobals->allFonts,sizeof(struct UserFont)*nrFonts);

   // now get the installed fonts
   DmListDatabases(&loadFonts, &i, 0);
   return 1;
}

void releaseAllFonts()
{
   int i;
   struct UserFont *uf;
   for (i=nrFonts-1; i >= 0; i--)
   {
      if ((uf=&vmGlobals->allFonts[i])->recH)
      {
         if (IS_UNICODE_FONT(uf))
            destroyUff(uf);
         else
         {
            //DmReleaseRecord(vmGlobals->allFonts[i].dmRef, 0, 0);
            xfree(vmGlobals->allFonts[i].recH);
            if (uf->recIndex == 0) DmCloseDatabase(uf->dmRef);
         }
      }
   }
   if (vmGlobals->allFonts) xfree(vmGlobals->allFonts);
}
// waba/fx/Font static void fontCreate(Font obj)
void FontCreate(Var stack[]) // (Lwaba/fx/Font;)V
{
   WObject obj = stack[0].obj; // class object
   int bold = WOBJ_FontStyle(obj) == Font_BOLD;
   int large = WOBJ_FontSize(obj) >= 14;
   UtfString fontName;
   int index;
   char *c;

   //static char *sufixes[4] = {"Small","SmallBold","Large","LargeBold"};
   char *sufix = 0;

   fontName = (*vmGlobals->stringToUtf)(WOBJ_FontName(obj), STU_USE_STATIC | STU_NULL_TERMINATE);
   c = xstrstr(fontName.str,"Small");
   if (c != NULL && c != fontName.str)
      *c = 0;
   c = xstrstr(fontName.str,"Large");
   if (c != NULL && c != fontName.str)
      *c = 0;

   if (fontName.len > 3 && (fontName.str[1] == fontName.str[2]) && (fontName.str[0] == 'U' || fontName.str[1] == 'u') && (fontName.str[1] == 'F' || fontName.str[1] == 'f')) // is it uff?
      sufix = (vmGlobals->screenWidth==160)?"_L":"_H";
   else
      switch (large*2+bold)
      {
         case 0: sufix = "Small";     break;
         case 1: sufix = "SmallBold"; break;
         case 2: sufix = "Large";     break;
         case 3: sufix = "LargeBold"; break;
      }
   // first, try to load the font itself using the sw font pattern
   index = getUserFontIndex(fontName.str,sufix);
   if (index == -1)
   {
      // ok, the pattern didnt work, use the font name given
      if (!(xstrlen(fontName.str) > 2 && fontName.str[0] == 'S' && fontName.str[1] == 'W')) // is it the internal ones? dont try the name itself
         index = getUserFontIndex(fontName.str,NULL);
      if (index == -1)
      {
         // ok, the font is not found, use the internal ones
         index = getUserFontIndex((vmGlobals->screenWidth >= 320)? "5SW" : "MSW", sufix);
         if (index == -1)
            index = bold?1:0; // internal ones not found (maybe large in palmOS 2.0), use the default small font.
         // change the font name
         WOBJ_FontName(obj) = (*vmGlobals->createString)(vmGlobals->allFonts[index].name);
      }
   }
   WOBJ_FontUserFont(obj) = vmGlobals->allFonts+index;
}
//////////////////////////////////////////////////////////////////////////
// waba/fx/FontMetrics static void fontMetricsCreate(FontMetrics obj)
void FontMetricsCreate(Var stack[]) // (Lwaba/fx/FontMetrics;)V
{
   WObject font, fontMetrics/*, surface*/; // guich@450_36: surface is ignored from now on
   struct UserFont *uf;

   fontMetrics = stack[0].obj;

   font = WOBJ_FontMetricsFont(fontMetrics);
   if (font == 0)
   {
      WOBJ_FontMetricsAscent(fontMetrics) = 0;
      WOBJ_FontMetricsDescent(fontMetrics) = 0;
      WOBJ_FontMetricsLeading(fontMetrics) = 0;
      return;
   }
   // surface is unused and user defined fonts are supported!
   uf = WOBJ_FontUserFont(font);
   if (uf == NULL || uf->fontP == NULL)
   {
      // MessageBox(NULL, (uf == NULL) ? TEXT("uf is NULL") : TEXT("uf->fontP is NULL"), TEXT("FontMetricsCreate"), MB_ICONEXCLAMATION);
      WOBJ_FontMetricsAscent(fontMetrics) = 10;
      WOBJ_FontMetricsDescent(fontMetrics) = 2;
      WOBJ_FontMetricsLeading(fontMetrics) = 1;
      return;
   }

   WOBJ_FontMetricsAscent(fontMetrics) = uf->fontP->ascent;
   WOBJ_FontMetricsDescent(fontMetrics) = uf->fontP->descent;
   WOBJ_FontMetricsLeading(fontMetrics) = uf->fontP->leading;
}
//////////////////////////////////////////////////////////////////////////
// waba/fx/FontMetrics static int fontMetricsGetCharWidth(FontMetrics obj, char c)
void FontMetricsGetCharWidth(Var stack[]) // (Lwaba/fx/FontMetrics;C)I
{
   FontMetricsGetWidth(FM_CHARWIDTH, stack);
}
//////////////////////////////////////////////////////////////////////////
// waba/fx/FontMetrics static int fontMetricsGetTextWidth(FontMetrics obj, String s)
void FontMetricsGetTextWidth_String(Var stack[]) // (Lwaba/fx/FontMetrics;Ljava/lang/String;)I
{
   FontMetricsGetWidth(FM_STRINGWIDTH, stack);
}
//////////////////////////////////////////////////////////////////////////
// waba/fx/FontMetrics static int fontMetricsGetTextWidth(FontMetrics obj, char chars[], int start, int count)
void FontMetricsGetTextWidth(Var stack[]) // (Lwaba/fx/FontMetrics;[CII)I
{
   FontMetricsGetWidth(FM_CHARARRAYWIDTH, stack);
}

#endif // defined PALMOS ----------------------------------------------------------------------
