/*********************************************************************************
 *  SuperWaba Virtual Machine, version 5                                         *
 *  Copyright (C) 2000-2005 SuperWaba Ltda.                                      *
 *  All Rights Reserved                                                          *
 *                                                                               *
 *  This library and virtual machine is distributed in the hope that it will     *
 *  be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                         *
 *                                                                               *
 *  Please see the software license located at /license.txt for more details.    *
 *                                                                               *
 *  You should have received a copy of the License along with this software;     *
 *  if not, write to                                                             *
 *                                                                               *
 *     SuperWaba Ltda.                                                           *
 *     Rua Padre Leonel Franca, 480, sala 27A                                    *
 *     Instituto Genesis - PUC-RIO                                               *
 *     Rio de Janeiro / RJ - Brazil                                              *
 *     Cep: 22451-000                                                            *
 *     E-mail: licensing@superwaba.com.br                                        *
 *********************************************************************************/
/* $Id: SW_libDynam_c.h,v 1.11 2005/11/13 02:19:19 fdie Exp $ */

// Before including superwaba.h, define BUILDING_CUSTOM_LIB so that when
// we make calls to the library's routines from this file, we generate
// the correct dll import/export symbols.

/********************************************************************
 * Private variables
 ********************************************************************/

// This array is used to map a hash value to a corresponding native function.
// It must remain sorted by hash value because a binary search is performed
// to find a method by its hash value. There is a small chance of collision
// if two hashes match and if one occurs, the function name should be changed
// to avoid collision. To prevent users from creating invalid native methods
// that hash to a valid value, native methods could be restricted to a specific
// set of classes (this is probably not necessary since any verifier probably
// wouldn't allow native methods to get by anyway).
// guich: since i introduced a hash table, this table dont has anymore to stay sorted. :-))

// guich@102 - implements a hash table, whose key is a int32 and value can be anything.
// ps: HASHSIZE is computed in the NativeHashes program.

struct htElem
{
   uint32 key;
   NativeFunc value;
   struct htElem *next;
};

DECLARE_ARRAY(static struct htElem *, hashTable, HASHSIZE); /* The storage array for the methods */

DECLARE_VAR(struct VMGlobals *, vmGlobals); // set during the SWLibOpen call

// error classes are PalmOS specific
#if !defined(appErrorClass)
#define appErrorClass 0
#endif

/********************************************************************
 * Private macros
 ********************************************************************/

// If we're built for an ARM device, then these macros are used to setup/restore
// the ELF GOT table (Global Offset Table) register R8 for nested library calls.
// On ARM, we reserved R8 as the native lib GOT table pointer, thus we never change
// the VM GOT table pointed by R10.
// On PalmOS 5, R9 is used by the operating system and cannot be changed neither.
// So, to compile your ARM native lib never forget these gcc settings:
// -fPIC -msoft-float -ffixed-r9 -ffixed-r10 -mpic-register=r8 -msingle-pic-base -march=armv4t

#if defined(PALMOS) && defined(__arm__)
// attention! We cannot use such an expression 'register void *got asm("r8")' 
// because we cannot forbid optimization.
#define EnterLibrary_(refnum) \
    register void *saved_got; \
    asm volatile ("mov %0, r8\n" : "=r" (saved_got) :); \
    asm volatile ("mov r8, %0\n" : : "r" (refnum));
#define ExitLibrary_() \
    asm volatile ("mov r8,%0" : : "r" (saved_got));
#else
#define EnterLibrary_(refnum)
#define ExitLibrary_()
#endif

#define htHash(key) (unsigned)((key) % HASHSIZE)

/********************************************************************
 * Private functions
 ********************************************************************/

/*************************************************************/
/* return the htElem found, else NULL if key doesn't exist   */
static struct htElem *htGet(uint32 key)
{
   struct htElem *np;
   int h = htHash(key);
   for (np = GLOBAL_VAR(hashTable)[h]; np; np = np->next)
   {
      if (np->key == key)
         return np;
   }
   return NULL;
}

/*************************************************************/
/* return struct containing value or NULL if unable to store */
static struct htElem *htPut(uint32 key, NativeFunc value)
{
   struct htElem *np;
   unsigned hashval;
   if ((np = htGet(key)) == NULL) // duplicated keys not allowed
   {
      np = (struct htElem *) VM_CALL(xmalloc) (sizeof(struct htElem));
      if (np != NULL)
      {
          hashval = htHash(key);
          /* point to last w/ this hash */
          np->next = GLOBAL_VAR(hashTable)[hashval];
          np->key = key;
          np->value = value;
          /* attach to hashtab array */
          GLOBAL_VAR(hashTable)[hashval] = np;
      }
   }
   else
   {
      VM_CALL(debug)(VM_CALL(int2str)("**** duplicated key in hashtable ****",key));
   }
   return np;
}

/*************************************************************/
static void disposeHashtable()
{
   // dispose the hashtable
   int i;
   for (i = 0; i < HASHSIZE; i++)
   {
      struct htElem *np;
      np = GLOBAL_VAR(hashTable)[i];
      while (np)
      {
         struct htElem *next;
         next = np->next;
         xfree(np);
         np = next;
      }
      GLOBAL_VAR(hashTable)[i] = NULL;
   }
}

/*************************************************************/

Err SWLibOpen(LibRefnum inRefnum, void *inGlobalsPtr)
{
   ClassHook *libClassHooks;
   int i;
   EnterLibrary_(inRefnum)

   // 1. associate the vmGlobals struct with the given one
   GLOBAL_VAR(vmGlobals) = (struct VMGlobals *)inGlobalsPtr;
   _LOG_INFO((LOG_CAT_NLIB, "SWLibOpen %d %x", inRefnum, inGlobalsPtr));

   // 2. install our hooks
   libClassHooks = getClassHooks();
   if (libClassHooks)
   {
#if defined(__SYMBIAN32__) || (defined(PALMOS) && defined(__arm__))
      uint16 libref = GLOBAL_VAR(vmGlobals)->tempLibInsertIndex; // guich@551_8: pass it in a var inside vmGlobals instead of hacking inRefnum
#else
      uint16 libref = inRefnum;
#endif
      int32 hookLibCount = VM_CALL(arrayCount)((char *)libClassHooks,sizeof(ClassHook));
      int32 insertIndex;
      VM_VAR(classHooks) = (ClassHook *)VM_CALL(expandArray)((char *)VM_VAR(classHooks),sizeof(ClassHook),hookLibCount,&insertIndex);
      if (insertIndex == -1) return SWErrMemory;
      for (i =0; i < hookLibCount; i++)
         libClassHooks[i].libRefnum = libref;

      xmemmove (VM_VAR(classHooks)+insertIndex,libClassHooks, hookLibCount*sizeof(ClassHook));
   }
   buildNativeFuncArray();
   for (i=0; GLOBAL_VAR(nativeHashes)[i]; i++)
   {
      htPut(GLOBAL_VAR(nativeHashes)[i], GLOBAL_VAR(nativeFuncs)[i]);
   }
   ExitLibrary_()
   return 0;
}

/*************************************************************/
Err SWLibClose(_unused_ LibRefnum inRefnum)
{
   EnterLibrary_(inRefnum)
   _LOG_INFO((LOG_CAT_NLIB, "SWLibClose %d", inRefnum));
   onClose();
   disposeHashtable();
   ExitLibrary_()
   return 0;
}

/*************************************************************/
Err SWGetNativeFunc(_unused_ LibRefnum inRefnum, int32 hash, NativeFunc *func)
{
   Err err = 0;
   struct htElem *elem;

   EnterLibrary_(inRefnum)
   _LOG_INFO((LOG_CAT_NLIB, "SWGetNativeFunc %d", inRefnum));
   elem = htGet(hash);
   if (elem && elem->value)
      *func = elem->value;
   else
      err = SWErrMethodNotFound;
   ExitLibrary_()
   return err;
}

/*************************************************************/
Err SWInvokeNativeFunc(_unused_ LibRefnum inRefnum, NativeFunc func, Var stack[])
{
   EnterLibrary_(inRefnum)
   _LOG_INFO((LOG_CAT_NLIB, "SWInvokeNativeFunc %d", inRefnum));
   func(stack);
   ExitLibrary_()
   return 0;
}

/*************************************************************/
Err SWInvokeNative(_unused_ LibRefnum inRefnum, int32 hash, Var stack[])
{
   Err err = 0;
   struct htElem *elem;

   EnterLibrary_(inRefnum)
   _LOG_INFO((LOG_CAT_NLIB, "SWInvokeNative %d", inRefnum));
   elem = htGet(hash);

   if (elem && elem->value)
   {
      elem->value(stack);
   }
   else
   {
      VM_CALL(debug)(VM_CALL(int2str)("SWInvokeNative coud'nt find method with hash ",hash));
      err = SWErrMethodNotFound;
   }
   ExitLibrary_()
   return err;
}

/*************************************************************/
Err SWCallPublic(_unused_ LibRefnum inRefnum, int index, uint32 *ret, uint32 *args)
{
   Err err;
   EnterLibrary_(inRefnum)
   _LOG_INFO((LOG_CAT_NLIB, "SWCallPublic %d", inRefnum));
   err = callPublic(index, ret, args);
   ExitLibrary_()
   return err;
}

/*************************************************************/
Err SWHandleEvent(_unused_ LibRefnum inRefnum, EventPtr eventP)
{
   Err err;
   EnterLibrary_(inRefnum)
   _LOG_INFO((LOG_CAT_NLIB, "SWHandleEvent %d", inRefnum));
   err = handleEvent(eventP);
   ExitLibrary_()
   return err;
}
